import pytest
import numpy as np


from genlm.control.potential import Potential
from genlm.control.sampler.sequence import SMC, SequenceModel
from genlm.control.sampler.token import DirectTokenSampler

from hypothesis import strategies as st, settings, given
from conftest import (
    weighted_set,
    weighted_sequence,
    double_weighted_sequence,
    WeightedSet,
)


@pytest.fixture
def default_unit_sampler():
    sequences = ["a", "b", "c"]
    weights = [1, 2, 3]
    p = WeightedSet(sequences, weights)
    return DirectTokenSampler(p)


@pytest.mark.asyncio
@settings(deadline=None)
@given(weighted_set(weighted_sequence))
async def test_importance(S):
    sequences, weights = zip(*S)

    p = WeightedSet(sequences, weights)
    unit_sampler = DirectTokenSampler(p)

    n_particles = 100
    sampler = SMC(unit_sampler)

    sequences = await sampler(n_particles=n_particles, ess_threshold=0, max_tokens=10)
    assert len(sequences) == n_particles
    assert np.isclose(sequences.log_ml, np.log(sum(weights)), atol=1e-3, rtol=1e-5)


@pytest.mark.asyncio
@settings(deadline=None)
@given(weighted_set(double_weighted_sequence))
async def test_importance_with_critic(S):
    sequences, weights1, weights2 = zip(*S)

    p = WeightedSet(sequences, weights1)
    unit_sampler = DirectTokenSampler(p)
    critic = WeightedSet(sequences, weights2)

    n_particles = 10
    sampler = SMC(unit_sampler, critic=critic)
    sequences = await sampler(n_particles=n_particles, ess_threshold=0, max_tokens=10)

    logeps = await p.prefix([])
    for seq, logw in sequences:
        logZ = sum([(await p.logw_next(seq[:n])).sum() for n in range(len(seq))])
        assert np.isclose(logw, logZ + logeps + await critic.score(seq))


@pytest.mark.asyncio
@settings(deadline=None)
@given(weighted_set(weighted_sequence), st.floats(min_value=0, max_value=1))
async def test_smc(S, ess_threshold):
    sequences, weights = zip(*S)

    p = WeightedSet(sequences, weights)
    unit_sampler = DirectTokenSampler(p)

    n_particles = 100
    sampler = SMC(unit_sampler)

    sequences = await sampler(
        n_particles=n_particles, ess_threshold=ess_threshold, max_tokens=10
    )
    assert len(sequences) == n_particles
    assert np.isclose(sequences.log_ml, np.log(sum(weights)), atol=1e-3, rtol=1e-5)


@pytest.mark.asyncio
@settings(deadline=None)
@given(st.floats(min_value=0, max_value=1))
async def test_smc_with_critic(ess_threshold):
    seqs = ["0", "00", "1"]
    weights1 = [3.0, 2.0, 1.0]
    weights2 = [1.0, 2.0, 3.0]

    p = WeightedSet(seqs, weights1)
    unit_sampler = DirectTokenSampler(p)
    critic = WeightedSet(seqs, weights2)

    n_particles = 500
    sampler = SMC(unit_sampler, critic=critic)

    sequences = await sampler(
        n_particles=n_particles, ess_threshold=ess_threshold, max_tokens=10
    )

    intersection_ws = [w1 * w2 for w1, w2 in zip(weights1, weights2)]
    assert len(sequences) == n_particles
    assert np.isclose(
        np.exp(sequences.log_ml), sum(intersection_ws), atol=0.5, rtol=0.05
    )


@st.composite
def smc_params(draw, item_sampler, max_seq_len=5, max_size=5):
    S = draw(weighted_set(item_sampler, max_seq_len, max_size))
    stop_point = draw(st.integers(min_value=1, max_value=max_seq_len))
    return S, stop_point


@pytest.mark.asyncio
@settings(deadline=None)
@given(smc_params(double_weighted_sequence))
async def test_smc_weights(params):
    S, stop_point = params
    sequences, weights1, weights2 = zip(*S)

    p = WeightedSet(sequences, weights1)
    unit_sampler = DirectTokenSampler(p)
    critic = WeightedSet(sequences, weights2)

    n_particles = 10
    sampler = SMC(unit_sampler, critic=critic)

    sequences = await sampler(
        n_particles=n_particles,
        ess_threshold=0,  # don't resample since that would reset weights
        max_tokens=stop_point,
    )

    logeps = await p.prefix([])
    for seq, logw in sequences:
        logZ = sum([(await p.logw_next(seq[:n])).sum() for n in range(len(seq))])
        twist = await critic.score(seq)
        assert np.isclose(logw, logZ + logeps + twist)


@pytest.mark.asyncio
async def test_sequence_model_invalid_start_weight():
    class MockPotential(Potential):
        async def prefix(self, context):
            if not context:
                return -np.inf
            return 0

        async def complete(self, context):
            return 0

    unit_sampler = DirectTokenSampler(MockPotential([0]))
    seq_model = SequenceModel(unit_sampler)
    with pytest.raises(ValueError, match="Start weight.*"):
        await seq_model.start()


def test_sequence_model_str_for_serialization(default_unit_sampler):
    SequenceModel(default_unit_sampler).string_for_serialization()



import numpy as np
from genlm.grammar import Float, Log
from arsenal.maths import logsumexp

from genlm.control.constant import EndOfSequence


class LazyWeights:
    """
    A class to represent weights in a lazy manner, allowing for efficient operations
    on potentially large weight arrays without immediate materialization.

    Attributes:
        weights (np.ndarray): The weights associated with the tokens.
        encode (dict): A mapping from tokens to their corresponding indices in the weights array.
        decode (list): A list of tokens corresponding to the weights.
        is_log (bool): A flag indicating whether the weights are in log space.
    """

    def __init__(self, weights, encode, decode, log=True):
        """
        Initialize the LazyWeights instance.

        Args:
            weights (np.ndarray): The weights associated with the tokens.
            encode (dict): A mapping from tokens to their corresponding indices in the weights array.
            decode (list): A list of tokens corresponding to the weights.
            log (bool, optional): Indicates if the weights are in log space. Defaults to True.

        Raises:
            AssertionError: If the lengths of weights and decode or encode do not match.
        """
        assert len(weights) == len(decode)
        assert len(encode) == len(decode)

        self.weights = weights
        self.encode = encode
        self.decode = decode
        self.is_log = log

    def __getitem__(self, token):
        """
        Retrieve the weight for a given token.

        Args:
            token (Any): The token for which to retrieve the weight.

        Returns:
            (float): The weight of the token, or -inf/0 if the token is not found.
        """
        if token not in self.encode:
            return float("-inf") if self.is_log else 0
        return self.weights[self.encode[token]]

    def __len__(self):
        return len(self.weights)

    def __array__(self):
        raise NotImplementedError(
            "LazyWeights cannot be converted to a numpy array. "
            "If you want to combine multiple LazyWeights, use their weights attribute directly."
        )

    def keys(self):
        """Return the list of tokens (keys) in the vocabulary."""
        return self.decode

    def values(self):
        """Return the weights associated with the tokens."""
        return self.weights

    def items(self):
        """Return a zip of tokens and weights."""
        return zip(self.keys(), self.values())

    def normalize(self):
        """
        Normalize the weights.

        Normalization is performed using log-space arithmetic when weights are logarithmic,
        or standard arithmetic otherwise.

        Returns:
            (LazyWeights): A new LazyWeights instance with normalized weights.
        """
        if self.is_log:
            return self.spawn(self.weights - logsumexp(self.weights))
        else:
            return self.spawn(self.weights / np.sum(self.weights))

    def exp(self):
        """
        Exponentiate the weights. This operation can only be performed when weights are in log space.

        Returns:
            (LazyWeights): A new LazyWeights instance with exponentiated weights.

        Raises:
            AssertionError: If the weights are not in log space.
        """
        assert self.is_log, "Weights must be in log space to exponentiate"
        return self.spawn(np.exp(self.weights), log=False)

    def log(self):
        """
        Take the logarithm of the weights. This operation can only be performed when weights are in regular space.

        Returns:
            (LazyWeights): A new LazyWeights instance with logarithmic weights.

        Raises:
            AssertionError: If the weights are already in log space.
        """
        assert not self.is_log, "Weights must be in regular space to take the logarithm"
        return self.spawn(np.log(self.weights), log=True)

    def sum(self):
        """
        Sum the weights.

        Summation is performed using log-space arithmetic when weights are logarithmic,
        or standard arithmetic otherwise.

        Returns:
            (float): The sum of the weights, either in log space or regular space.
        """
        if self.is_log:
            return logsumexp(self.weights)
        else:
            return np.sum(self.weights)

    def spawn(self, new_weights, log=None):
        """
        Create a new LazyWeights instance over the same vocabulary with new weights.

        Args:
            new_weights (np.ndarray): The new weights for the LazyWeights instance.
            log (bool, optional): Indicates if the new weights are in log space. Defaults to None.

        Returns:
            (LazyWeights): A new LazyWeights instance.
        """
        if log is None:
            log = self.is_log
        return LazyWeights(
            weights=new_weights, encode=self.encode, decode=self.decode, log=log
        )

    def materialize(self, top=None):
        """
        Materialize the weights into a chart.

        Args:
            top (int, optional): The number of top weights to materialize. Defaults to None.

        Returns:
            (Chart): A chart representation of the weights.
        """
        weights = self.weights
        if top is not None:
            top_ws = weights.argsort()[-int(top) :]
        else:
            top_ws = weights.argsort()

        semiring = Log if self.is_log else Float

        chart = semiring.chart()
        for i in reversed(top_ws):
            chart[self.decode[i]] = weights[i]

        return chart

    def __repr__(self):
        return repr(self.materialize())

    def assert_equal(self, other, **kwargs):
        """
        Assert that two LazyWeights instances are equal.

        This method asserts that the two LazyWeights instances have the same vocabulary
        (in identical order) and that their weights are numerically close.

        Args:
            other (LazyWeights): The other LazyWeights instance to compare.
            **kwargs (dict): Additional arguments for np.testing.assert_allclose (e.g., rtol, atol).
        """
        assert self.decode == other.decode
        np.testing.assert_allclose(self.weights, other.weights, **kwargs)

    def assert_equal_unordered(self, other, **kwargs):
        """
        Assert that two LazyWeights instances are equal, ignoring vocabularyorder.

        Args:
            other (LazyWeights): The other LazyWeights instance to compare.
            **kwargs (dict): Additional arguments for np.isclose (e.g., rtol, atol).
        """
        assert set(self.decode) == set(other.decode), "keys do not match"

        for x in self.decode:
            have, want = self[x], other[x]
            assert np.isclose(have, want, **kwargs), f"{x}: {have} != {want}"


def load_trie(V, backend=None, **kwargs):
    """
    Load a TokenCharacterTrie.

    Args:
        V (list): The vocabulary.
        backend (str, optional): The backend to use for trie construction. Defaults to None.
        **kwargs (dict): Additional arguments for the trie construction.

    Returns:
        (TokenCharacterTrie): A trie instance.
    """
    import torch

    if backend is None:
        backend = "parallel" if torch.cuda.is_available() else "sequential"

    if backend == "parallel":
        from genlm.backend.trie import ParallelTokenCharacterTrie

        return ParallelTokenCharacterTrie(V, **kwargs)
    else:
        from genlm.backend.trie import TokenCharacterTrie

        return TokenCharacterTrie(V, **kwargs)


def load_async_trie(V, backend=None, **kwargs):
    """
    Load an AsyncTokenCharacterTrie. This is a TokenCharacterTrie that
    automatically batches weight_sum and weight_max requests.

    Args:
        V (list): The vocabulary.
        backend (str, optional): The backend to use for trie construction. Defaults to None.
        **kwargs (dict): Additional arguments for the trie construction.

    Returns:
        (AsyncTokenCharacterTrie): An async trie instance.
    """
    from genlm.backend.trie import AsyncTokenCharacterTrie

    return AsyncTokenCharacterTrie(load_trie(V, backend, **kwargs))


def fast_sample_logprobs(logprobs: np.ndarray, size: int = 1) -> np.ndarray:
    """Sample indices from an array of log probabilities using the Gumbel-max trick.

    Args:
        logprobs: Array of log probabilities
        size (int): Number of samples to draw

    Returns:
        (np.ndarray): Array of sampled indices

    Note:
        This is much faster than np.random.choice for large arrays since it avoids
        normalizing probabilities and uses vectorized operations.
    """
    noise = -np.log(-np.log(np.random.random((size, len(logprobs)))))
    return (logprobs + noise).argmax(axis=1)


def fast_sample_lazyweights(lazyweights):
    """Sample a token from a LazyWeights instance using the Gumbel-max trick.

    Args:
        lazyweights (LazyWeights): A LazyWeights instance

    Returns:
        (Any): Sampled token
    """
    assert lazyweights.is_log
    token_id = fast_sample_logprobs(lazyweights.weights, size=1)[0]
    return lazyweights.decode[token_id]


def escape(x):
    if isinstance(x, EndOfSequence):
        return repr(x)
    elif isinstance(x, int):  # assume its a byte
        x = bytes([x])
    if isinstance(x, bytes):
        y = repr(x)[2:-1]
    else:
        y = repr(x)[1:-1]
    return y.replace(" ", "␣")



import pytest
import asyncio
import numpy as np
from genlm.control.typing import Atomic
from genlm.control.potential.base import Potential, EOS


class SimplePotential(Potential):
    async def complete(self, context):
        return -float(len(context))  # Scoring based on length

    async def prefix(self, context):
        return -0.5 * float(len(context))  # Different scoring for prefixes


@pytest.fixture
def potential():
    return SimplePotential([b"a", b"b", b"c"])


def test_token_type(potential):
    assert potential.token_type == Atomic(bytes)


@pytest.mark.asyncio
async def test_score(potential):
    context = [b"b", b"c"]

    have = await potential.score(context)
    want = await potential.prefix(context)
    assert want == -1.0
    assert have == want

    have = await potential.score(context + [EOS])
    want = await potential.complete(context)
    assert want == -2.0
    assert have == want

    have = await potential.score([])
    want = await potential.prefix([])
    assert want == 0.0
    assert have == want


@pytest.mark.asyncio
async def test_logw_next(potential):
    context = [b"b", b"c"]
    have = (await potential.logw_next(context)).materialize()
    for token in potential.vocab_eos:
        want = await potential.score(context + [token]) - await potential.prefix(
            context
        )
        assert have[token] == want


@pytest.mark.asyncio
async def test_batch_score(potential):
    seq1 = [b"a"]
    seq2 = [b"a", b"b"]
    seq3 = [b"a", b"b", EOS]

    have = await potential.batch_score([seq1, seq2, seq3])
    want = await asyncio.gather(
        potential.score(seq1), potential.score(seq2), potential.score(seq3)
    )

    np.testing.assert_array_equal(have, want)
    np.testing.assert_array_equal(have, [-0.5, -1.0, -2.0])


@pytest.mark.asyncio
async def test_batch_logw_next(potential):
    seq1 = [b"a"]
    seq2 = [b"b", b"c"]

    haves = await potential.batch_logw_next([seq1, seq2])
    wants = await asyncio.gather(potential.logw_next(seq1), potential.logw_next(seq2))

    for want, have in zip(haves, wants):
        np.testing.assert_array_equal(have.weights, want.weights)


@pytest.mark.asyncio
async def test_empty(potential):
    with pytest.raises(ValueError):
        await potential.batch_logw_next([])

    with pytest.raises(ValueError):
        await potential.batch_score([])

    with pytest.raises(ValueError):
        await potential.batch_prefix([])

    with pytest.raises(ValueError):
        await potential.batch_complete([])


@pytest.mark.asyncio
async def test_properties(potential):
    await potential.assert_logw_next_consistency([b"b", b"c"], verbosity=1)
    await potential.assert_autoreg_fact([b"b", b"c"], verbosity=1)
    await potential.assert_batch_consistency([[b"b", b"c"], [b"a"]], verbosity=1)


def test_initialization_errors():
    # Test empty vocabulary
    with pytest.raises(ValueError, match="vocabulary cannot be empty"):
        SimplePotential([])

    # Test invalid token_type
    with pytest.raises(ValueError, match="token_type must be a TokenType"):
        SimplePotential([b"a"], token_type="not a token type")

    # Test wrong token types in vocabulary
    wrong_type = Atomic(str)  # Using str instead of bytes
    with pytest.raises(TypeError, match="Tokens in vocabulary must be of type"):
        SimplePotential([b"a", b"b"], token_type=wrong_type)

    # Test invalid EOS type
    with pytest.raises(ValueError, match="EOS must be an instance of EndOfSequence"):
        SimplePotential([b"a"], eos="not an EndOfSequence")

    # Test duplicate tokens
    with pytest.raises(ValueError, match="Duplicate token.*found in vocabulary"):
        SimplePotential([b"a", b"a"])


@pytest.mark.asyncio
async def test_zero_weight_context():
    class ZeroWeightPotential(SimplePotential):
        async def prefix(self, context):
            return float("-inf")

    potential = ZeroWeightPotential([b"a", b"b"])
    with pytest.raises(ValueError, match="Context.*has weight zero under `prefix`."):
        await potential.logw_next([b"a"])


def test_spawn_not_implemented():
    potential = SimplePotential([b"a"])
    with pytest.raises(
        NotImplementedError,
        match="Potential.spawn\\(\\) must be implemented by subclasses.",
    ):
        potential.spawn()



from genlm.control.potential.stateful import StatefulPotential, ParticleState
from abc import ABC, abstractmethod
from enum import Enum, auto
import asyncio


class Responses(Enum):
    INCOMPLETE = auto()
    COMPLETE = auto()
    ERROR = auto()


class UniqueIdentifier:
    def __init__(self, name):
        self.__name = name

    def __repr__(self):
        return self.__name


PING_TOKEN = UniqueIdentifier("PING_TOKEN")
SHUTDOWN_TOKEN = UniqueIdentifier("SHUTDOWN_TOKEN")


# This should be an async generator really but async generators
# are fundamentally broken. See https://peps.python.org/pep-0789/
# I kept running into problems with this during implementation, so
# ended up finding it easier to just hand roll implementations of
# this rather than trying to use yield based generators.
class AsyncSource(ABC):
    @abstractmethod
    async def more(self): ...


class Chunks(AsyncSource):
    def __init__(self, running_in_task):
        self.running_in_task = running_in_task
        self.__first = True

    async def more(self):
        await self.running_in_task.responses.put(
            (self.running_in_task.last_message, Responses.INCOMPLETE)
        )
        (
            self.running_in_task.last_message,
            chunk,
        ) = await self.running_in_task.incoming_data.get()
        if chunk is SHUTDOWN_TOKEN:
            raise StopAsyncIteration()
        return chunk


class RunningInTask:
    def __init__(self, function):
        self.incoming_data = asyncio.Queue()
        self.responses = asyncio.Queue()
        self.last_message = None
        self.running = False
        self.complete = False
        self.function = function

    async def run(self):
        assert not self.running
        try:
            self.running = True
            self.last_message, chunk = await self.incoming_data.get()
            assert chunk == PING_TOKEN
            chunks = Chunks(self)
            result = await self.function(chunks)
        except Exception as e:
            await self.responses.put((self.last_message, Responses.ERROR, e))
        else:
            await self.responses.put((self.last_message, Responses.COMPLETE, result))
        finally:
            self.running = False
            self.complete = True


# This is sortof insane, but asyncio will get *very* upset with you if your task
# objects are garbage collected before they're complete. This keeps a set of them
# around until they're completed.
KEEP_ALIVE_SET = set()


class AsyncStreamingState(ParticleState):
    def __init__(self, owner):
        super().__init__(owner)
        self.__token = 0
        self.__background = None
        self.__score = 0.0
        self.diagnostics = {}

    def __new_token(self):
        self.__token += 1
        return self.__token

    async def __initialize_background(self):
        if self.__background is None:
            self.__background = RunningInTask(self.owner.calculate_score_from_stream)
            self.__background_task = asyncio.create_task(self.__background.run())
            await self.__send_message(PING_TOKEN)
            KEEP_ALIVE_SET.add(self.__background_task)
            self.__background_task.add_done_callback(KEEP_ALIVE_SET.discard)
        assert self.__background is not None

    async def impl_update_context(self, incremental_context):
        await self.__initialize_background()
        finish = False
        if incremental_context and incremental_context[-1] == self.owner.eos:
            finish = True
            incremental_context = incremental_context[:-1]
        bytes(incremental_context)
        await self.__send_message(incremental_context)
        if finish:
            await self.finish()

    async def impl_finish(self):
        await self.__initialize_background()
        await self.shutdown()

    @property
    def current_score(self):
        return self.__score

    async def __send_message(self, message):
        if self.__background.complete:
            return
        token = (self.__new_token(), message)
        await self.__background.incoming_data.put((token, message))

        (
            response_token,
            response_type,
            *payload,
        ) = await self.__background.responses.get()

        assert token == response_token
        match response_type:
            case Responses.INCOMPLETE:
                pass
            case Responses.COMPLETE:
                self.__score = payload[0] or 0.0
            case Responses.ERROR:
                self.diagnostics["error"] = payload
                self.__score = -float("inf")

    async def shutdown(self):
        if self.__background is not None:
            await self.__send_message(SHUTDOWN_TOKEN)


class AsyncStreamingPotential(StatefulPotential, ABC):
    def __init__(self, vocabulary, token_type=None, eos=None):
        super().__init__(
            vocabulary=vocabulary,
            token_type=token_type,
            eos=eos,
            state_class=AsyncStreamingState,
        )

    @abstractmethod
    async def calculate_score_from_stream(self, stream: AsyncSource) -> float: ...



import pytest
import numpy as np
from genlm.control.sampler.sequence import Sequences, EndOfSequence, EOS


def test_initialization():
    sequences = Sequences(
        contexts=[[b"a"], [b"b"]],
        log_weights=[np.log(0.4), np.log(0.6)],
    )
    assert sequences.size == 2
    assert np.isclose(np.exp(sequences.log_total), 1.0)  # weights sum to 1
    assert len(sequences) == 2


def test_initialization_validation():
    # Test mismatched lengths
    with pytest.raises(AssertionError):
        Sequences(contexts=[[b"a"]], log_weights=[0.0, 0.0])


def test_posterior():
    # Test posterior without EOS filtering
    sequences = Sequences(
        contexts=[
            [b"hello"],  # No EOS
            [b"world", EndOfSequence()],
        ],
        log_weights=[np.log(0.4), np.log(0.6)],
    )
    posterior = sequences.posterior
    assert len(posterior) == 2
    assert np.isclose(posterior[tuple([b"hello"])], 0.4)
    assert np.isclose(posterior[tuple([b"world", EndOfSequence()])], 0.6)


def test_normalized_weights():
    sequences = Sequences(
        contexts=[[b"a"], [b"b"]],
        log_weights=[np.log(3), np.log(7)],
    )
    weights = sequences.normalized_weights
    assert np.allclose(weights, [0.3, 0.7])
    assert np.isclose(np.sum(weights), 1.0)


def test_iteration_and_indexing():
    contexts = [[b"a"], [b"b"]]
    log_weights = [np.log(0.3), np.log(0.7)]
    sequences = Sequences(contexts=contexts, log_weights=log_weights)

    # Test __iter__
    for i, (ctx, weight) in enumerate(sequences):
        assert ctx == contexts[i]
        assert weight == log_weights[i]

    # Test __getitem__
    assert sequences[0] == (contexts[0], log_weights[0])
    assert sequences[1] == (contexts[1], log_weights[1])


def test_effective_sample_size():
    # Test equal weights (maximum ESS)
    sequences = Sequences(
        contexts=[[b"a"], [b"b"], [b"c"]],
        log_weights=[0.0, 0.0, 0.0],  # equal weights
    )
    assert np.isclose(sequences.ess, 3.0)  # ESS should equal number of particles

    # Test completely unbalanced weights (minimum ESS)
    sequences = Sequences(
        contexts=[[b"a"], [b"b"], [b"c"]],
        log_weights=[
            np.log(1.0),
            float("-inf"),
            float("-inf"),
        ],  # one particle has all weight
    )
    assert np.isclose(sequences.ess, 1.0)


def test_log_ml_calculation():
    # Test log marginal likelihood calculation
    sequences = Sequences(
        contexts=[[b"a"], [b"b"]],
        log_weights=[np.log(0.3), np.log(0.7)],
    )
    assert np.isfinite(sequences.log_ml)
    assert sequences.log_ml <= sequences.log_total


def test_empty_sequences():
    sequences = Sequences(contexts=[], log_weights=[])
    assert sequences.size == 0
    assert len(sequences.posterior) == 0
    assert len(sequences.decoded_posterior) == 0


def test_posterior_normalization():
    # Test that posterior probabilities sum to 1
    sequences = Sequences(
        contexts=[
            [b"hello", EndOfSequence()],
            [b"world", EndOfSequence()],
            [b"test", EndOfSequence()],
        ],
        log_weights=[np.log(2), np.log(5), np.log(3)],
    )
    posterior = sequences.posterior
    assert np.isclose(sum(posterior.values()), 1.0)


def test_string_representation():
    sequences = Sequences(contexts=[[b"test", EndOfSequence()]], log_weights=[0.0])
    # Test that string representation doesn't raise errors
    str(sequences)
    repr(sequences)


def test_decoded_posterior_basic_sequence():
    # Simple case with one valid UTF-8 sequence
    sequences = Sequences(contexts=[[b"hello", EndOfSequence()]], log_weights=[0.0])
    posterior = sequences.decoded_posterior
    assert len(posterior) == 1
    assert posterior["hello"] == 1.0


def test_decoded_posterior_multiple_sequences():
    # Multiple different valid sequences
    sequences = Sequences(
        contexts=[[b"hello", EndOfSequence()], [b"world", EndOfSequence()]],
        log_weights=[np.log(0.7), np.log(0.3)],
    )
    posterior = sequences.decoded_posterior
    assert len(posterior) == 2
    assert np.isclose(posterior["hello"], 0.7)
    assert np.isclose(posterior["world"], 0.3)


def test_duplicate_sequences():
    # Test that duplicate sequences have their probabilities summed
    sequences = Sequences(
        contexts=[
            [b"hello", EndOfSequence()],
            [b"hello", EndOfSequence()],
            [b"world", EndOfSequence()],
        ],
        log_weights=[np.log(4), np.log(4), np.log(2)],
    )
    posterior = sequences.decoded_posterior
    assert len(posterior) == 2
    assert np.isclose(posterior["hello"], 0.8)
    assert np.isclose(posterior["world"], 0.2)


def test_no_eos_sequences():
    # Test when no sequences end with EOS
    sequences = Sequences(
        contexts=[[b"hello"], [b"world"]],
        log_weights=[np.log(0.6), np.log(0.4)],
    )
    posterior = sequences.decoded_posterior
    assert len(posterior) == 0


def test_mixed_eos_and_non_eos():
    # Test mixture of EOS and non-EOS sequences
    sequences = Sequences(
        contexts=[
            [b"hello", EndOfSequence()],
            [b"world"],  # No EOS
            [b"test", EndOfSequence()],
        ],
        log_weights=[np.log(5), np.log(2), np.log(3)],
    )
    posterior = sequences.decoded_posterior
    assert len(posterior) == 2
    # Note: weights should be renormalized after filtering
    total_weight = 5 + 3
    assert np.isclose(posterior["hello"], 5 / total_weight)
    assert np.isclose(posterior["test"], 3 / total_weight)


def test_invalid_utf8_sequences():
    # Test handling of invalid UTF-8 sequences
    invalid_bytes = bytes([0xFF, 0xFF])  # Invalid UTF-8
    sequences = Sequences(
        contexts=[
            [b"hello", EndOfSequence()],
            [invalid_bytes, EndOfSequence()],
            [b"world", EndOfSequence()],
        ],
        log_weights=[np.log(4), np.log(2), np.log(4)],
    )
    posterior = sequences.decoded_posterior
    assert len(posterior) == 2
    total_weight = 4 + 4
    assert np.isclose(posterior["hello"], 4 / total_weight)
    assert np.isclose(posterior["world"], 4 / total_weight)


def test_empty_sequence_with_eos():
    # Test sequence that's just EOS
    sequences = Sequences(contexts=[[EndOfSequence()]], log_weights=[0.0])
    posterior = sequences.decoded_posterior
    assert len(posterior) == 1
    assert posterior[""] == 1.0


def test_multi_byte_utf8():
    # Test with multi-byte UTF-8 characters
    sequences = Sequences(
        contexts=[
            ["🌟".encode("utf-8"), EndOfSequence()],
            ["こんにちは".encode("utf-8"), EndOfSequence()],
        ],
        log_weights=[np.log(3), np.log(7)],
    )
    posterior = sequences.decoded_posterior
    assert len(posterior) == 2
    assert np.isclose(posterior["🌟"], 0.3)
    assert np.isclose(posterior["こんにちは"], 0.7)


def test_all_negative_infinity_weights():
    # Test handling of case where all weights are -inf
    sequences = Sequences(
        contexts=[[b"hello", EndOfSequence()], [b"world", EndOfSequence()]],
        log_weights=[-np.inf, -np.inf],
    )

    # Check all the derived quantities
    assert sequences.log_total == float("-inf")
    assert sequences.log_ml == float("-inf")
    assert np.all(np.isneginf(sequences.log_normalized_weights))
    assert sequences.log_ess == float("-inf")
    assert sequences.ess == 0.0

    # Check that posterior methods handle this case
    assert len(sequences.posterior) == 2
    assert len(sequences.decoded_posterior) == 2


def test_shows():
    sequences = Sequences(
        contexts=[[b"a", b"b", b"c", EOS], [b"a", b"b", b"d"]],
        log_weights=[np.log(1), np.log(9)],
    )
    sequences.show()
    repr(sequences)
    sequences._repr_html_()
    str(sequences)



import pytest
from genlm.control.typing import (
    Atomic,
    Sequence,
    infer_type,
    infer_vocabulary_type,
)


def test_atomic_type_check():
    int_type = Atomic(int)
    assert int_type.check(42)
    assert not int_type.check("42")

    str_type = Atomic(str)
    assert str_type.check("hello")
    assert not str_type.check(b"hello")

    bytes_type = Atomic(bytes)
    assert bytes_type.check(b"hello")
    assert not bytes_type.check("hello")


def test_sequence_type_check():
    int_seq = Sequence(Atomic(int))
    assert int_seq.check([1, 2, 3])
    assert not int_seq.check([1, "2", 3])
    assert not int_seq.check(42)

    nested_int_seq = Sequence(Sequence(Atomic(int)))
    assert nested_int_seq.check([[1, 2], [3, 4]])
    assert not nested_int_seq.check([[1, 2], 3])

    bytes_seq = Sequence(Atomic(bytes))
    assert bytes_seq.check([b"hello", b"world"])
    assert not bytes_seq.check(["hello", "world"])


def test_atomic_inference():
    assert infer_type(42) == Atomic(int)
    assert infer_type("hello") == Atomic(str)
    assert infer_type(b"hello") == Atomic(bytes)
    assert infer_type(3.14) == Atomic(float)
    assert infer_type(True) == Atomic(bool)


def test_sequence_inference():
    assert infer_type([1, 2, 3]) == Sequence(Atomic(int))
    assert infer_type(["a", "b"]) == Sequence(Atomic(str))
    assert infer_type([[1, 2], [3, 4]]) == Sequence(Sequence(Atomic(int)))
    assert infer_type([b"AB", b"CD"]) == Sequence(Atomic(bytes))


def test_empty_sequence_error():
    with pytest.raises(ValueError):
        infer_type([])


def test_inconsistent_sequence_error():
    with pytest.raises(ValueError):
        infer_type([1, "2", 3])


def test_is_iterable_of():
    assert Sequence(Atomic(int)).is_iterable_of(Atomic(int))
    assert Sequence(Atomic(str)).is_iterable_of(Atomic(str))
    assert not Sequence(Atomic(int)).is_iterable_of(Atomic(str))

    assert Atomic(bytes).is_iterable_of(Atomic(int))
    assert Atomic(str).is_iterable_of(Atomic(str))

    assert not Atomic(int).is_iterable_of(Atomic(int))
    assert not Atomic(bytes).is_iterable_of(Atomic(str))
    assert not Atomic(str).is_iterable_of(Atomic(int))

    nested_seq = Sequence(Sequence(Atomic(int)))
    assert nested_seq.is_iterable_of(Sequence(Atomic(int)))
    assert not nested_seq.is_iterable_of(Atomic(int))


def test_vocabulary_type_inference():
    """Test the infer_vocabulary_type function"""
    assert infer_vocabulary_type([1, 2, 3]) == Atomic(int)
    assert infer_vocabulary_type(["a", "b"]) == Atomic(str)
    assert infer_vocabulary_type([[1, 2], [3, 4]]) == Sequence(Atomic(int))

    # Test empty vocabulary
    with pytest.raises(ValueError):
        infer_vocabulary_type([])

    # Test inconsistent types
    with pytest.raises(ValueError):
        infer_vocabulary_type([1, "2", 3])


def test_atomic_convert():
    int_type = Atomic(int)
    assert int_type.convert(42) == 42
    assert int_type.convert("42") == 42
    repr(int_type)


def test_sequence_convert():
    int_seq = Sequence(Atomic(int))
    assert int_seq.convert([1, 2, 3]) == (1, 2, 3)
    assert int_seq.convert(["1", "2", "3"]) == (1, 2, 3)
    repr(int_seq)



import numpy as np
from genlm.grammar import Float
from arsenal.maths import logsumexp
from functools import cached_property
from dataclasses import dataclass
from arsenal import colors

from llamppl import Model
from llamppl import smc_standard

from genlm.control.potential import Potential
from genlm.control.constant import EOS, EndOfSequence
from genlm.control.sampler.token import TokenSampler
from genlm.control.util import escape


class SMC:
    """This class implements sequential Monte Carlo (SMC) inference for controlled text generation.
    The generation process works as follows:

    1. Token Sampling: At each step, the `unit_sampler` is used to extend each particle (candidate sequence)
       by sampling a new token. This grows all sequences by one token at a time. The sampler also outputs
       an importance weight with each extension to correct for the myopic nature of token-by-token sampling.

    2. Critic Evaluation: If a `critic` is provided, it scores the updated sequences (via it's `score` method),
       reweighting the particles based on how well they satisfy the constraints encoded by the critic.

    3. Resampling: When the effective sample size (ESS) falls below the threshold,
       particles are resampled according to their weights. This helps focus computation
       on more promising sequences.

    4. Termination: The process continues until either:\n
        - All sequences reach an end-of-sequence (EOS) token\n
        - The maximum token length is reached

    If a critic is provided, the resulting sequences are properly weighted with respect to the product of the unit sampler's
    target potential and the critic potential (`unit_sampler.target * critic`). If a critic is not provided,
    the resulting sequences are weighted with respect to the unit sampler's target potential.

    Args:
        unit_sampler (TokenSampler): The sampler that generates tokens.
        critic (Potential, optional): A potential function that guides the generation process
            by scoring candidate sequences. Must have the same token type as the unit_sampler.

    Raises:
        ValueError: If unit_sampler is not a TokenSampler, if critic is not a Potential,
            or if the token types of unit_sampler and critic don't match.
    """

    def __init__(self, unit_sampler, critic=None):
        if not isinstance(unit_sampler, TokenSampler):
            raise ValueError("`unit_sampler` must be a TokenSampler")

        if critic:
            if not isinstance(critic, Potential):
                raise ValueError("`critic` must be a Potential")
            if not unit_sampler.token_type == critic.token_type:
                raise ValueError(
                    "`critic` must have the same token type as the `unit_sampler`. "
                    f"Got {unit_sampler.token_type} and {critic.token_type}."
                    + (
                        "\nMaybe you forgot to coerce the critic to the token type of the unit sampler? See `Coerce`."
                        if unit_sampler.token_type.is_iterable_of(critic.token_type)
                        else ""
                    )
                )

        self.unit_sampler = unit_sampler
        self.critic = critic

    async def __call__(
        self,
        n_particles,
        ess_threshold,
        max_tokens,
        verbosity=0,
        json_path=None,
        **kwargs,
    ):
        """Generate sequences using sequential Monte Carlo inference.

        Args:
            n_particles (int): Number of particles (candidate sequences) to maintain during
                generation. Higher values provide better exploration but require more
                computation.
            ess_threshold (float): Effective sample size threshold for resampling,
                expressed as a fraction of the number of particles. When ESS falls below
                this value, particles are resampled according to their weights. Should be between 0 and 1.
                Higher values lead to more frequent resampling. Note that when ess_threshold = 0,
                the critic is only applied at the end of the generation (if it is provided).
            max_tokens (int): Maximum number of tokens to generate per sequence. Generation
                may terminate earlier if all sequences reach an EOS token.
            verbosity (int, optional): Verbosity level for the SMC algorithm. 0 is silent, 1 prints the
                particles at each step. Default is 0.
            json_path (str, optional): JSON file path for saving a record of the inference run.
                This can be used in conjunction with the `InferenceVisualizer` to visualize the inference run.
            **kwargs (dict): Additional keyword arguments to pass to the SMC algorithm.
                See the `llamppl.inference.smc_standard` documentation for more details.

        Returns:
            (Sequences): A container holding the generated sequences, their importance weights, and
                other metadata from the generation process.
        """
        model = SequenceModel(
            unit_sampler=self.unit_sampler,
            critic=self.critic,
            max_tokens=max_tokens,
            verbosity=verbosity,
            twist_with_critic=ess_threshold > 0,
        )

        particles = await smc_standard(
            model=model,
            n_particles=n_particles,
            ess_threshold=ess_threshold,
            json_file=json_path,
            **kwargs,
        )

        return Sequences(*_unpack_particles(particles))

    async def cleanup(self):
        """Clean up resources used by the inference engine.

        This method should be called when the InferenceEngine is no longer needed.

        Example:
            ```python
            sampler = SequenceSampler(unit_sampler, critic)
            try:
                sequences = await sampler(n_particles=10, ess_threshold=0.5, max_tokens=20)
            finally:
                await sampler.cleanup()
            ```
        """
        await self.unit_sampler.cleanup()
        if self.critic:
            await self.critic.cleanup()


@dataclass
class Sequences:
    """Container for sequence samples with their weights and probabilities.

    Args:
        contexts (list): List of token sequences generated by the sampler.
        log_weights (list): Log importance weights for each sequence.

    Attributes:
        size (int): Number of sequences in the container.
        logp (float): Sum of log probabilities across all sequences.
        log_total (float): Log of the sum of importance weights.
        log_ml (float): Log marginal likelihood estimate.
        log_normalized_weights (list): Log weights normalized to sum to 1.
        log_ess (float): Log of the effective sample size.
        ess (float): Effective sample size of the particle population.
    """

    contexts: list
    log_weights: list

    def __post_init__(self):
        assert len(self.contexts) == len(self.log_weights)

        if not isinstance(self.log_weights, np.ndarray):
            self.log_weights = np.array(self.log_weights)

        self.size = len(self.contexts)

        # Handle case where all weights are -inf
        if np.all(np.isneginf(self.log_weights)):
            self.log_total = float("-inf")
            self.log_ml = float("-inf")
            self.log_normalized_weights = np.full_like(self.log_weights, float("-inf"))
            self.log_ess = float("-inf")
            self.ess = 0.0
            return

        self.log_total = logsumexp(self.log_weights)
        max_weight = max(self.log_weights)
        self.log_ml = (
            np.log(np.mean(np.exp(self.log_weights - max_weight))) + max_weight
        )
        self.log_normalized_weights = self.log_weights - self.log_total
        self.log_ess = -logsumexp(2 * self.log_normalized_weights)
        self.ess = np.exp(self.log_ess)

    @cached_property
    def posterior(self):
        """Compute the estimated posterior distribution over sequences.

        The probability of a sequence corresponds to its normalized weight. The probabilities
        of duplicate sequences are summed.

        Returns:
            (Float.chart): A normalized chart mapping sequences to their posterior probabilities,
                sorted in descending order by probability.
        """
        posterior = Float.chart()
        for sequence, prob in zip(self.contexts, self.normalized_weights):
            posterior[tuple(sequence)] += prob
        return posterior.normalize().sort_descending()

    @cached_property
    def decoded_posterior(self):
        """Compute posterior distribution over completed UTF-8 decodable sequences.

        Filters for sequences that:\n
        1. End with an EndOfSequence token\n
        2. Can be decoded as UTF-8 strings

        The probability of each sequence corresponds to its normalized weight among completed and decodable sequences.
        Probabilities of duplicate sequences (after decoding) are summed.

        To obtain the posterior distribution over all byte sequences, use `self.posterior`.

        Returns:
            (Float.chart): A normalized chart mapping decoded string sequences to their
                posterior probabilities, sorted in descending order by probability.
                Only includes sequences that meet both filtering criteria.
        """
        posterior = Float.chart()
        for sequence, w in zip(self.contexts, np.exp(self.log_weights)):
            if sequence and isinstance(sequence[-1], EndOfSequence):
                try:
                    string_sequence = b"".join(sequence[:-1]).decode("utf-8")
                    posterior[string_sequence] += w
                except UnicodeDecodeError:
                    pass
        return posterior.normalize().sort_descending()

    @property
    def normalized_weights(self):
        """Return exponential of normalized log weights."""
        if np.all(np.isneginf(self.log_weights)):
            return np.full_like(self.log_weights, 0.0)
        return np.exp(self.log_normalized_weights)

    def __len__(self):
        return self.size

    def __iter__(self):
        return iter(zip(self.contexts, self.log_weights))

    def __getitem__(self, i):
        return self.contexts[i], self.log_weights[i]

    def __str__(self):
        return str(self.decoded_posterior)

    def _repr_html_(self):
        return self.decoded_posterior._repr_html_()

    def __repr__(self):
        return str(self.decoded_posterior)

    def show(self):
        for p in sorted(self, reverse=True):
            print(p)


class SequenceModel(Model):
    def __init__(
        self,
        unit_sampler,
        critic=None,
        max_tokens=float("inf"),
        verbosity=0,
        twist_with_critic=True,
    ):
        assert max_tokens > 0

        super().__init__()
        self.token_ctx = []
        self.unit_sampler = unit_sampler
        self.max_tokens = max_tokens
        self.critic = critic
        self.logp = 0
        self.verbosity = verbosity
        self.twist_with_critic = twist_with_critic

    async def start(self):
        start_w = await self.unit_sampler.start_weight()
        if start_w == float("-inf"):
            raise ValueError(
                "Start weight is -inf (log(0)). This is likely because a potential assigns zero weight to "
                "the empty sequence under `prefix`, which violates the potential contract."
            )
        self.score(start_w)

    async def step(self):
        unit = await self.call(self.unit_sampler)
        self.token_ctx.append(unit)

        inf_weight = self.weight == float("-inf")
        if inf_weight:
            if self.critic:
                assert self.twist_amount != float("-inf")
            self.finish()
            return

        if self.critic and self.twist_with_critic:
            twist_amt = await self.critic.score(self.token_ctx)
            if twist_amt != float("-inf"):
                self.twist(twist_amt)
            else:
                self.score(twist_amt)
                self.finish()
                return

        if self.verbosity > 0:
            print(self.__repr__())

        self.max_tokens -= 1
        if self.max_tokens == 0 or self.token_ctx[-1] is EOS:
            self.finish()
            if self.critic:
                if not self.twist_with_critic:
                    twist_amt = await self.critic.score(self.token_ctx)
                self.score(twist_amt)
            return

    def __repr__(self):
        return (
            f"{self.weight:.2f}:\t"
            + colors.magenta % "["
            + (colors.magenta % "|").join(escape(y) for y in self.token_ctx)
            + colors.magenta % "]"
        )

    def string_for_serialization(self):
        return "|".join(escape(y) for y in self.token_ctx)

    def immutable_properties(self):
        return set(["unit_sampler", "critic"])


def _unpack_particles(particles):
    contexts, logws = map(
        list,
        zip(
            *[
                (p.token_ctx, float("-inf") if np.isnan(p.weight) else p.weight)
                for p in particles
            ]
        ),
    )
    return contexts, logws



from dataclasses import dataclass
from collections.abc import Sequence as SequenceABC

from genlm.control.constant import EndOfSequence


@dataclass
class TokenType:
    """Base class representing the type of a token"""

    def check(self, value):
        """Check if a value matches this type"""
        raise NotImplementedError()  # pragma: no cover

    def is_iterable_of(self, element_type):
        """Check if this type can be interpreted as an iterable of element_type.

        Args:
            element_type (TokenType): The type to check if this is an iterable of

        Examples:
            >>> Sequence(Atomic(int)).is_iterable_of(Atomic(int))
            True
            >>> Atomic(bytes).is_iterable_of(Atomic(int))
            True
        """
        if isinstance(self, Sequence):
            return self.element_type == element_type

        if isinstance(self, Atomic):
            # Special cases for built-in iterables
            if (
                self.type is bytes
                and isinstance(element_type, Atomic)
                and element_type.type is int
            ):
                return True
            if (
                self.type is str
                and isinstance(element_type, Atomic)
                and element_type.type is str
            ):
                return True

        return False


@dataclass
class Atomic(TokenType):
    """Represents a simple type like int or str"""

    type: type  # The Python type (int, str, etc.)

    def check(self, value):
        return isinstance(value, self.type) or isinstance(value, EndOfSequence)

    def convert(self, value):
        return self.type(value)

    def __repr__(self):
        return f"Atomic({self.type.__name__})"


@dataclass
class Sequence(TokenType):
    """Represents a list/sequence of another type"""

    element_type: TokenType  # The type of elements in the sequence

    def check(self, value):
        return isinstance(value, (list, tuple)) and all(
            self.element_type.check(x) for x in value
        )

    def convert(self, value):
        return tuple(self.element_type.convert(x) for x in value)

    def __repr__(self):
        return f"Sequence({self.element_type!r})"


def infer_type(value):
    """Infer the TokenType from a value.

    Args:
        value (Any): A sample value to infer type from

    Returns:
        (TokenType): The inferred type

    Examples:
        >>> infer_type(42)
        Atomic(type=int)
        >>> infer_type([1, 2, 3])
        Sequence(element_type=Atomic(type=int))
        >>> infer_type([[1, 2], [3, 4]])
        Sequence(element_type=Sequence(element_type=Atomic(type=int)))
    """
    if isinstance(value, SequenceABC) and not isinstance(value, (bytes, str)):
        if not value:
            raise ValueError("Cannot infer type from empty sequence")
        element_type = infer_type(value[0])
        if not all(element_type.check(x) for x in value):
            raise ValueError("Inconsistent types in sequence")
        return Sequence(element_type)

    return Atomic(type(value))


def infer_vocabulary_type(vocabulary):
    """Infer the TokenType from a vocabulary.

    Args:
        vocabulary (List[Any]): A list of tokens to infer type from

    Returns:
        (TokenType): The inferred type

    Raises:
        ValueError: If vocabulary is empty or contains inconsistent types

    Examples:
        >>> infer_vocabulary_type([1, 2, 3])
        Atomic(type=int)
        >>> infer_vocabulary_type([[1, 2], [3, 4]])
        Sequence(element_type=Atomic(type=int))
    """
    if not vocabulary:
        raise ValueError("Cannot infer type from empty vocabulary")

    token_type = infer_type(vocabulary[0])
    if not all(token_type.check(x) for x in vocabulary):
        raise ValueError("Inconsistent types in vocabulary")

    return token_type



import pytest
import asyncio
import time
import numpy as np
from genlm.control.potential import Potential


class MockPotential(Potential):
    """Mock potential for testing with controlled delays"""

    def __init__(self):
        super().__init__(list(range(256)))
        self.delay = 0.1  # 100ms delay per operation

    async def complete(self, context):
        time.sleep(self.delay)
        return np.log(len(context))

    async def prefix(self, context):
        time.sleep(self.delay)
        return np.log(len(context) / 2)

    async def batch_complete(self, contexts):
        time.sleep(self.delay)  # Single delay for batch
        return np.array([np.log(len(context)) for context in contexts])

    async def batch_prefix(self, contexts):
        time.sleep(self.delay)  # Single delay for batch
        return np.array([np.log(len(context) / 2) for context in contexts])

    def spawn(self):
        return MockPotential()


@pytest.mark.asyncio
async def test_correctness():
    """Test that autobatched results match sequential results"""
    potential = MockPotential()
    autobatched = potential.to_autobatched()

    sequences = [b"hello", b"world", b"test", b"batch", b"foo"]

    want = await asyncio.gather(*(potential.complete(seq) for seq in sequences))
    have = await asyncio.gather(*(autobatched.complete(seq) for seq in sequences))
    assert want == have, [want, have]

    want = await asyncio.gather(*(potential.prefix(seq) for seq in sequences))
    have = await asyncio.gather(*(autobatched.prefix(seq) for seq in sequences))
    assert want == have, [want, have]

    want = await asyncio.gather(*(potential.score(seq) for seq in sequences))
    have = await asyncio.gather(*(autobatched.score(seq) for seq in sequences))
    assert want == have, [want, have]

    wants = await asyncio.gather(*(potential.logw_next(seq) for seq in sequences))
    haves = await asyncio.gather(*(autobatched.logw_next(seq) for seq in sequences))
    for have, want in zip(haves, wants):
        have.assert_equal(want)

    await autobatched.cleanup()


@pytest.mark.asyncio
async def test_batch_methods():
    """Test that batch methods return expected results (they shouldn't change)"""
    potential = MockPotential()
    autobatched = potential.to_autobatched()

    sequences = [b"hello", b"world", b"test", b"batch", b"foo"]

    want_complete = await potential.batch_complete(sequences)
    have_complete = await autobatched.batch_complete(sequences)
    np.testing.assert_array_equal(want_complete, have_complete)

    want_prefix = await potential.batch_prefix(sequences)
    have_prefix = await autobatched.batch_prefix(sequences)
    np.testing.assert_array_equal(want_prefix, have_prefix)

    want_score = await potential.batch_score(sequences)
    have_score = await autobatched.batch_score(sequences)
    np.testing.assert_array_equal(want_score, have_score)

    want_logw_next = await potential.batch_logw_next(sequences)
    have_logw_next = await autobatched.batch_logw_next(sequences)
    for have, want in zip(have_logw_next, want_logw_next):
        have.assert_equal(want)

    await autobatched.cleanup()


@pytest.mark.asyncio
async def test_performance():
    """Test that autobatched operations are faster than sequential"""
    potential = MockPotential()
    autobatched = potential.to_autobatched()

    sequences = [b"hello", b"world", b"test", b"batch", b"foo"]

    start = time.perf_counter()
    await asyncio.gather(*(potential.complete(seq) for seq in sequences))
    sequential_time = time.perf_counter() - start

    start = time.perf_counter()
    await asyncio.gather(*(autobatched.complete(seq) for seq in sequences))
    autobatched_time = time.perf_counter() - start

    print(sequential_time, autobatched_time)

    assert autobatched_time < sequential_time / 2

    await autobatched.cleanup()


@pytest.mark.asyncio
async def test_error_handling():
    """Test that errors in batch processing are properly propagated"""

    class ErrorPotential(MockPotential):
        async def batch_complete(self, contexts):
            raise ValueError("Test error")

    potential = ErrorPotential()
    autobatched = potential.to_autobatched()

    with pytest.raises(ValueError, match="Test error"):
        await autobatched.complete(b"test")

    await autobatched.cleanup()


@pytest.mark.asyncio
async def test_spawn_and_repr():
    """Test spawn method creates new instance and repr works correctly"""
    potential = MockPotential()
    autobatched = potential.to_autobatched()

    # Test spawn
    spawned = autobatched.spawn()
    assert isinstance(spawned, type(autobatched))
    assert spawned is not autobatched
    assert spawned.potential is not autobatched.potential

    # Test repr
    expected_repr = f"AutoBatchedPotential({potential!r})"
    assert repr(autobatched) == expected_repr

    await autobatched.cleanup()
    await spawned.cleanup()


@pytest.mark.asyncio
async def test_close_and_cleanup():
    """Test close() and cleanup() methods"""
    potential = MockPotential()
    autobatched = potential.to_autobatched()

    # Test that the background loop is running
    assert autobatched.background_loop.task is not None
    assert not autobatched.background_loop.task.done()

    # Test close()
    autobatched.background_loop.close()
    assert autobatched.background_loop.task is None

    # Test cleanup()
    autobatched = potential.to_autobatched()  # Create new instance
    await autobatched.cleanup()
    assert autobatched.background_loop.task is None


@pytest.mark.asyncio
async def test_del_cleanup():
    """Test __del__ cleanup"""
    potential = MockPotential()
    autobatched = potential.to_autobatched()

    # Get reference to background loop
    loop = autobatched.background_loop
    assert loop.task is not None

    # Delete the autobatched instance
    del autobatched

    # Verify the background loop was cleaned up
    assert loop.task is None



import re
import pytest
import graphviz
import numpy as np
from genlm.grammar import WFSA as BaseWFSA, Float, Log, Boolean
from genlm.control.potential.built_in import WFSA, BoolFSA
from hypothesis import strategies as st, given, settings


@pytest.fixture
def float_wfsa():
    """Creates a simple WFSA in float semiring"""
    m = BaseWFSA(Float)
    m.add_I(0, 1.0)
    m.add_arc(0, b"a"[0], 1, 2)
    m.add_arc(1, b"b"[0], 2, 1)
    m.add_arc(1, b"c"[0], 2, 1)
    m.add_arc(1, b"d"[0], 3, 1)  # dead end
    m.add_F(2, 1.0)
    return m


@pytest.fixture
def log_wfsa():
    """Creates a simple WFSA in float semiring"""
    m = BaseWFSA(Log)
    m.add_I(0, Log(0.0))
    m.add_arc(0, b"a"[0], 1, Log(0.0))
    m.add_arc(1, b"b"[0], 2, Log(np.log(0.6)))
    m.add_arc(1, b"c"[0], 2, Log(np.log(0.4)))
    m.add_arc(1, b"d"[0], 3, Log(-float("inf")))  # dead end
    m.add_F(2, Log(0.0))
    return m


@pytest.mark.asyncio
async def test_wfsa(float_wfsa):
    pot = WFSA(float_wfsa)

    log_weight = await pot.complete(b"ab")
    assert np.isclose(log_weight, np.log(2))

    log_weight = await pot.complete(b"ac")
    assert np.isclose(log_weight, np.log(2))

    log_weight = await pot.complete(b"a")
    assert log_weight == float("-inf")

    log_weight = await pot.prefix(b"a")
    assert np.isclose(log_weight, np.log(4))

    log_weight = await pot.prefix(b"c")
    assert log_weight == float("-inf")

    log_weight = await pot.prefix(b"ab")
    assert np.isclose(log_weight, np.log(2))

    await pot.assert_logw_next_consistency(b"a")
    await pot.assert_autoreg_fact(b"a")

    await pot.assert_logw_next_consistency(b"")
    await pot.assert_autoreg_fact(b"")

    await pot.assert_batch_consistency([b"", b"ab", b"ac"])


@pytest.mark.asyncio
async def test_wfsa_regex():
    pot = WFSA.from_regex("a(b|c)")

    log_weight = await pot.complete(b"ab")
    assert np.isclose(log_weight, np.log(0.5))

    log_weight = await pot.complete(b"ac")
    assert np.isclose(log_weight, np.log(0.5))

    log_weight = await pot.complete(b"a")
    assert log_weight == float("-inf")

    log_weight = await pot.prefix(b"a")
    assert np.isclose(log_weight, 0)

    log_weight = await pot.prefix(b"c")
    assert log_weight == float("-inf")

    log_weight = await pot.prefix(b"ab")
    assert np.isclose(log_weight, np.log(0.5))

    await pot.assert_logw_next_consistency(b"a")
    await pot.assert_autoreg_fact(b"a")

    await pot.assert_logw_next_consistency(b"")
    await pot.assert_autoreg_fact(b"")

    await pot.assert_batch_consistency([b"", b"ab", b"ac"])


@pytest.mark.asyncio
async def test_bool_fsa(float_wfsa):
    pot = BoolFSA(float_wfsa)

    log_weight = await pot.complete(b"ab")
    assert log_weight == 0

    log_weight = await pot.complete(b"ac")
    assert log_weight == 0

    log_weight = await pot.complete(b"a")
    assert log_weight == float("-inf")

    log_weight = await pot.prefix(b"a")
    assert log_weight == 0

    log_weight = await pot.prefix(b"c")
    assert log_weight == float("-inf")

    log_weight = await pot.prefix(b"ab")
    assert log_weight == 0

    await pot.assert_logw_next_consistency(b"a")
    await pot.assert_autoreg_fact(b"a")

    await pot.assert_logw_next_consistency(b"")
    await pot.assert_autoreg_fact(b"")

    await pot.assert_batch_consistency([b"", b"ab", b"ac"])


@pytest.mark.asyncio
async def test_wfsa_long_ctx():
    # Test that we don't underflow when the context is long.
    pot = BoolFSA.from_regex(r".*")
    long_ctx = b"a" * 1000

    log_weight = await pot.complete(long_ctx)
    assert log_weight == 0

    log_weight = await pot.prefix(long_ctx)
    assert log_weight == 0


@st.composite
def regex_pattern(draw, max_depth=3):
    """Composite strategy to generate nested regex patterns"""

    def pattern_strategy(depth):
        if depth <= 0:
            # Base case: single escaped character
            char = draw(st.characters(blacklist_categories=("Cs",)))
            return re.escape(char)

        # Choose which type of pattern to generate
        pattern_type = draw(
            st.sampled_from(
                [
                    "simple",
                    "alternation",
                    "concatenation",
                    "optional",
                    "kleene",
                    "plus",
                    "quantified",
                ]
            )
        )

        if pattern_type == "simple":
            return pattern_strategy(0)

        # Generate sub-pattern(s)
        if pattern_type in ("alternation", "concatenation"):
            num_patterns = draw(st.integers(min_value=2, max_value=3))
            patterns = [pattern_strategy(depth - 1) for _ in range(num_patterns)]

            if pattern_type == "alternation":
                return f"({'|'.join(patterns)})"
            else:  # concatenation
                return f"({''.join(patterns)})"

        # Single sub-pattern with operator
        sub_pattern = pattern_strategy(depth - 1)
        if pattern_type == "optional":
            return f"({sub_pattern})?"
        elif pattern_type == "kleene":
            return f"({sub_pattern})*"
        elif pattern_type == "plus":
            return f"({sub_pattern})+"
        else:  # quantified
            quantifier = draw(st.sampled_from(["+", "*", "?", "{1,3}"]))
            return f"({sub_pattern}){quantifier}"

    return pattern_strategy(max_depth)


@pytest.mark.asyncio
@settings(deadline=None)
@given(regex_pattern(max_depth=3), st.data())
async def test_bool_fsa_with_generated_regex(pattern, data):
    """Test that BoolFSA accepts strings that match its regex pattern"""
    pot = BoolFSA.from_regex(pattern)

    matching_str = data.draw(st.from_regex(pattern, fullmatch=True))
    byte_string = matching_str.encode("utf-8")

    log_weight = await pot.complete(byte_string)
    assert log_weight == 0, [matching_str, pattern]

    for prefix in range(len(byte_string)):
        log_weight = await pot.prefix(byte_string[:prefix])
        assert log_weight == 0, [matching_str, byte_string[:prefix]]


def test_wfsa_init_wrong_semiring():
    # Test initialization with unsupported semiring
    wfsa = BaseWFSA(Boolean)  # TODO: support this semiring
    with pytest.raises(ValueError, match="Unsupported semiring"):
        WFSA(wfsa=wfsa)


def test_wfsa_init_float_conversion(log_wfsa):
    # Test that Float semiring is converted to Log
    pot = WFSA(wfsa=log_wfsa)
    assert pot.wfsa.R is Log


def test_wfsa_init_log_no_conversion(log_wfsa):
    # Test that Log semiring is not converted
    pot = WFSA(wfsa=log_wfsa)
    assert pot.wfsa.R is Log
    assert pot.wfsa is log_wfsa


def test_wfsa_repr(log_wfsa):
    pot = WFSA(wfsa=log_wfsa)
    repr(pot)

    try:
        pot._repr_svg_()
    except graphviz.backend.execute.ExecutableNotFound:
        pytest.skip("Graphviz not installed")


def test_bool_fsa_repr(log_wfsa):
    pot = BoolFSA(wfsa=log_wfsa)
    repr(pot)

    try:
        pot._repr_svg_()
    except graphviz.backend.execute.ExecutableNotFound:
        pytest.skip("Graphviz not installed")


def test_wfsa_spawn(log_wfsa):
    pot = WFSA(wfsa=log_wfsa)
    spawned = pot.spawn()
    assert isinstance(spawned, WFSA)


def test_wfsa_clear_cache(log_wfsa):
    pot = WFSA(wfsa=log_wfsa)
    pot.clear_cache()
    assert len(pot.cache) == 1
    assert () in pot.cache


@pytest.mark.asyncio
async def test_zero_weight_context():
    pot = WFSA.from_regex(r"a")
    with pytest.raises(ValueError, match="Context.*has zero weight."):
        await pot.logw_next(b"b")

    pot = BoolFSA.from_regex(r"a")
    with pytest.raises(ValueError, match="Context.*has zero weight."):
        await pot.logw_next(b"b")



import html
import numpy as np
from arsenal import Integerizer, colors
from arsenal.maths import sample, logsumexp
from graphviz import Digraph
from genlm.grammar import Float
from genlm.control.potential import Potential
from hypothesis import strategies as st


class MockPotential(Potential):
    def __init__(self, vocab, next_token_logws):
        self.next_token_logws = np.array(next_token_logws)
        super().__init__(vocab)

    def _logw(self, context):
        return sum([self.next_token_logws[self.lookup[i]] for i in context])

    async def prefix(self, context):
        return self._logw(context)

    async def complete(self, context):
        return self._logw(context) + self.next_token_logws[-1]

    async def logw_next(self, context):
        return self.make_lazy_weights(self.next_token_logws)


@st.composite
def mock_vocab(draw):
    item_strategy = draw(
        st.sampled_from(
            (
                st.text(min_size=1),
                st.binary(min_size=1),
            )
        )
    )

    # Sample vocabulary of iterables.
    vocab = draw(st.lists(item_strategy, min_size=1, max_size=10, unique=True))
    return vocab


@st.composite
def mock_vocab_and_ws(draw, max_w=1e3):
    vocab = draw(mock_vocab())
    ws = draw(
        st.lists(
            st.floats(1e-5, max_w),
            min_size=len(vocab) + 1,
            max_size=len(vocab) + 1,
        )
    )
    return vocab, ws


@st.composite
def mock_params(draw, max_w=1e3):
    iter_vocab, iter_next_token_ws = draw(mock_vocab_and_ws())

    # Sample context from iter_vocab
    context = draw(st.lists(st.sampled_from(iter_vocab), min_size=0, max_size=10))

    return (iter_vocab, iter_next_token_ws, context)


@st.composite
def iter_item_params(draw, max_iter_w=1e3, max_item_w=1e3):
    iter_vocab, iter_next_token_ws, context = draw(mock_params(max_iter_w))

    item_vocab = set()
    for items in iter_vocab:
        item_vocab.update(items)
    item_vocab = list(item_vocab)

    # Sample weights over item vocabulary and EOS.
    item_next_token_ws = draw(
        st.lists(
            st.floats(1e-5, max_item_w),
            min_size=len(item_vocab) + 1,
            max_size=len(item_vocab) + 1,
        )
    )

    return (iter_vocab, iter_next_token_ws, item_vocab, item_next_token_ws, context)


class WeightedSet(Potential):
    def __init__(self, sequences, weights):
        self.complete_logws = {
            tuple(seq): np.log(w) if w != 0 else float("-inf")
            for seq, w in zip(sequences, weights)
        }

        prefix_ws = {}
        for seq, w in zip(sequences, weights):
            for i in range(0, len(seq) + 1):
                prefix = tuple(seq[:i])
                if prefix not in prefix_ws:
                    prefix_ws[prefix] = 0.0
                prefix_ws[prefix] += w

        self.prefix_log_ws = {
            prefix: np.log(w) if w != 0 else float("-inf")
            for prefix, w in prefix_ws.items()
        }
        total_weight = sum(weights)
        assert np.isclose(
            self.prefix_log_ws[()],
            np.log(total_weight) if total_weight != 0 else float("-inf"),
        )

        super().__init__(list(set(t for seq in sequences for t in seq)))

    async def complete(self, context):
        return self.complete_logws.get(tuple(context), float("-inf"))

    async def prefix(self, context):
        return self.prefix_log_ws.get(tuple(context), float("-inf"))


@st.composite
def weighted_sequence(draw, max_seq_len=5):
    sequence = draw(st.text(min_size=1, max_size=max_seq_len))
    weight = draw(st.floats(min_value=1e-3, max_value=1e3))
    return sequence, weight


@st.composite
def double_weighted_sequence(draw, max_seq_len=5):
    # We use the second weight as the weight assigned to the sequence
    # by the critic.
    sequence = draw(st.text(min_size=1, max_size=max_seq_len))
    weight1 = draw(st.floats(min_value=1e-3, max_value=1e3))
    weight2 = draw(st.floats(min_value=0, max_value=1e3))
    return sequence, weight1, weight2


@st.composite
def weighted_set(draw, item_sampler, max_seq_len=5, max_size=5):
    return draw(
        st.lists(
            item_sampler(max_seq_len),
            min_size=1,
            max_size=max_size,
            unique_by=lambda x: x[0],
        )
    )


def separate_keys_vals(x):
    from genlm.control.util import LazyWeights

    if isinstance(x, LazyWeights):
        return x.keys(), x.values()
    elif isinstance(x, np.ndarray):
        return range(len(x)), x
    else:
        return list(x.keys()), np.array(list(x.values()))


class Tracer:
    """
    This class lazily materializes the probability tree of a generative process by program tracing.
    """

    def __init__(self):
        self.root = Node(idx=-1, mass=1.0, parent=None)
        self.cur = None

    def __call__(self, p, context=None):
        "Sample an action while updating the trace cursor and tree data structure."

        keys, p = separate_keys_vals(p)
        cur = self.cur

        if cur.child_masses is None:
            cur.child_masses = cur.mass * p
            cur.context = context

        if context != cur.context:
            print(colors.light.red % "ERROR: trace divergence detected:")
            print(colors.light.red % "trace context:", self.cur.context)
            print(colors.light.red % "calling context:", context)
            raise ValueError((p, cur))

        a = cur.sample()
        if a not in cur.active_children:
            cur.active_children[a] = Node(
                idx=a,
                mass=cur.child_masses[a],
                parent=cur,
                token=keys[a],
            )
        self.cur = cur.active_children[a]
        return keys[a]


class Node:
    __slots__ = (
        "idx",
        "mass",
        "parent",
        "token",
        "child_masses",
        "active_children",
        "context",
        "_mass",
    )

    def __init__(
        self,
        idx,
        mass,
        parent,
        token=None,
        child_masses=None,
        context=None,
    ):
        self.idx = idx
        self.mass = mass
        self.parent = parent
        self.token = token  # used for visualization
        self.child_masses = child_masses
        self.active_children = {}
        self.context = context
        self._mass = mass  # bookkeeping: remember the original mass

    def sample(self):
        return sample(self.child_masses)

    def p_next(self):
        return Float.chart((a, c.mass / self.mass) for a, c in self.children.items())

    # TODO: untested
    def sample_path(self):
        curr = self
        path = []
        P = 1
        while True:
            p = curr.p_next()
            a = curr.sample()
            P *= p[a]
            curr = curr.children[a]
            if not curr.children:
                break
            path.append(a)
        return (P, path, curr)

    def update(self):
        # TODO: Fennwick tree alternative, sumheap
        # TODO: optimize this by subtracting from masses, instead of resumming
        "Restore the invariant that self.mass = sum children mass."
        if self.parent is not None:
            self.parent.child_masses[self.idx] = self.mass
            self.parent.mass = np.sum(self.parent.child_masses)
            self.parent.update()

    def graphviz(
        self,
        fmt_edge=lambda x, a, y: f"{html.escape(str(a))}/{y._mass / x._mass:.2g}",
        # fmt_node=lambda x: ' ',
        fmt_node=lambda x: (
            f"{x.mass}/{x._mass:.2g}" if x.mass > 0 else f"{x._mass:.2g}"
        ),
    ):
        "Create a graphviz instance for this subtree"
        g = Digraph(
            graph_attr=dict(rankdir="LR"),
            node_attr=dict(
                fontname="Monospace",
                fontsize="10",
                height=".05",
                width=".05",
                margin="0.055,0.042",
            ),
            edge_attr=dict(arrowsize="0.3", fontname="Monospace", fontsize="9"),
        )
        f = Integerizer()
        xs = set()
        q = [self]
        while q:
            x = q.pop()
            xs.add(x)
            if x.child_masses is None:
                continue
            for a, y in x.active_children.items():
                a = y.token if y.token is not None else a
                g.edge(str(f(x)), str(f(y)), label=f"{fmt_edge(x, a, y)}")
                q.append(y)
        for x in xs:
            if x.child_masses is not None:
                g.node(str(f(x)), label=str(fmt_node(x)), shape="box")
            else:
                g.node(str(f(x)), label=str(fmt_node(x)), shape="box", fillcolor="gray")
        return g

    def downstream_nodes(self):
        q = [self]
        while q:
            x = q.pop()
            yield x
            if x.child_masses is None:
                continue
            for y in x.active_children.values():
                q.append(y)


class TraceSWOR(Tracer):
    """
    Sampling without replacement 🤝 Program tracing.
    """

    def __enter__(self):
        self.cur = self.root

    def __exit__(self, *args):
        self.cur.mass = 0  # we will never sample this node again.
        self.cur.update()  # update invariants

    def _repr_svg_(self):
        return self.root.graphviz()._repr_image_svg_xml()

    def sixel_render(self):
        try:
            from sixel import converter
            import sys
            from io import BytesIO

            c = converter.SixelConverter(
                BytesIO(self.root.graphviz()._repr_image_png())
            )
            c.write(sys.stdout)
        except ImportError:
            import warnings

            warnings.warn("Install imgcat or sixel to enable rendering.")
            print(self)

    def __repr__(self):
        return self.root.graphviz().source


async def trace_swor(sampler, context):
    tracer = TraceSWOR()
    logP = sampler.target.alloc_logws()
    while tracer.root.mass > 0:
        with tracer:
            token, logw, logp = await sampler.sample(context, draw=tracer)
            token_id = sampler.target.lookup[token]
            logP[token_id] = logsumexp([logP[token_id], logw + logp])

    return sampler.target.make_lazy_weights(logP)


async def trace_swor_set(sampler, context):
    tracer = TraceSWOR()
    logws = sampler.target.alloc_logws()
    while tracer.root.mass > 0:
        with tracer:
            set_logws, logp = await sampler.sample_set(context, draw=tracer)
            for token_id, logw in enumerate(set_logws.weights):
                if logw == float("-inf"):
                    continue
                logws[token_id] = logsumexp([logws[token_id], logw + logp])

    return sampler.target.make_lazy_weights(logws)



import pytest
import json
import socket
import tempfile
import time
import requests
from pathlib import Path
from genlm.control.viz import InferenceVisualizer


@pytest.fixture
def mocker(request):
    """Fixture to provide mocker."""
    return request.getfixturevalue("mocker")


def is_port_in_use(port):
    """Check if a port is in use."""
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        return s.connect_ex(("localhost", port)) == 0


@pytest.fixture
def viz():
    """Fixture that provides a visualizer and ensures cleanup."""
    visualizer = InferenceVisualizer()
    yield visualizer
    visualizer.shutdown_server()


@pytest.fixture
def test_data():
    return [
        {
            "step": 1,
            "mode": "init",
            "particles": [
                {
                    "contents": "<<<>>>b'h'",
                    "logweight": "-11.892930183943907",
                    "weight_incr": "-11.892930183943907",
                }
            ],
        },
    ]


def test_server_starts_on_default_port():
    """Test that server starts on the default port (8000)."""
    assert not is_port_in_use(8000)
    viz = InferenceVisualizer()
    try:
        assert is_port_in_use(8000)
    finally:
        viz.shutdown_server()


def test_server_uses_specified_port():
    """Test that server uses the specified port."""
    assert not is_port_in_use(8001)
    viz = InferenceVisualizer(port=8001)
    try:
        assert is_port_in_use(8001)
        assert not is_port_in_use(8000)
    finally:
        viz.shutdown_server()


def test_visualization_with_custom_dir(test_data):
    """Test visualization with a custom serve directory."""
    with tempfile.TemporaryDirectory() as serve_dir:
        viz = InferenceVisualizer(serve_dir=serve_dir)
        try:
            # Create a test JSON file in the serve directory
            json_path = Path(serve_dir) / "test.json"
            with open(json_path, "w") as f:
                json.dump(test_data, f)

            # Should be able to visualize immediately
            response = requests.get(f"http://localhost:8000/{json_path.name}")
            assert response.status_code == 200
            assert response.json() == test_data
        finally:
            viz.shutdown_server()


def test_visualization_with_external_file(test_data):
    """Test visualization with a file outside the serve directory."""
    viz = InferenceVisualizer()
    try:
        # Create a test JSON file in a different directory
        with tempfile.NamedTemporaryFile(suffix=".json", mode="w", delete=False) as f:
            json.dump(test_data, f)
            json_path = Path(f.name)

        # Should copy file and make it available
        viz.visualize(json_path)
        response = requests.get(f"http://localhost:8000/{json_path.name}")
        assert response.status_code == 200
        assert response.json() == test_data

    finally:
        viz.shutdown_server()
        json_path.unlink()


def test_server_cleanup():
    """Test that server cleanup works correctly."""
    viz = InferenceVisualizer()
    temp_dir = viz._serve_dir
    assert temp_dir.exists()

    viz.shutdown_server()
    time.sleep(0.5)  # Give the server a moment to fully shut down

    # Verify server is shut down and temp directory is cleaned up
    assert not is_port_in_use(8000)
    assert not temp_dir.exists()


def test_port_in_use():
    """Test that appropriate error is raised when port is in use."""
    viz1 = InferenceVisualizer(port=8002)
    try:
        with pytest.raises(OSError, match="Port.*already in use"):
            InferenceVisualizer(port=8002)
    finally:
        viz1.shutdown_server()


def test_html_file_request():
    """Test that HTML file requests are handled correctly."""
    viz = InferenceVisualizer()
    try:
        response = requests.get("http://localhost:8000/smc.html")
        assert response.status_code == 200
        assert "text/html" in response.headers["content-type"]
    finally:
        viz.shutdown_server()


def test_server_not_running():
    """Test error when server is not running."""
    viz = InferenceVisualizer()
    viz.shutdown_server()
    with pytest.raises(RuntimeError, match="Server is not running"):
        viz.visualize("test.json")


def test_file_not_found():
    """Test error when JSON file doesn't exist."""
    viz = InferenceVisualizer()
    try:
        with pytest.raises(FileNotFoundError, match="JSON file not found"):
            viz.visualize("nonexistent.json")
    finally:
        viz.shutdown_server()


def test_auto_open_browser(mocker):
    """Test auto-opening browser functionality."""
    mock_open = mocker.patch("webbrowser.open")
    viz = InferenceVisualizer()
    try:
        with tempfile.NamedTemporaryFile(suffix=".json", mode="w") as f:
            json.dump([], f)
            f.flush()
            viz.visualize(f.name, auto_open=True)
        mock_open.assert_called_once()
    finally:
        viz.shutdown_server()


def test_other_oserror(mocker):
    """Test handling of OSError other than port in use."""
    with mocker.patch(
        "socketserver.TCPServer.server_bind",
        side_effect=OSError(99, "Some other error"),
    ):
        with pytest.raises(OSError, match="Some other error"):
            InferenceVisualizer()



import torch
import warnings
from typing import NamedTuple
from genlm.control.potential.base import Potential


def load_model_by_name(name, backend, **kwargs):
    if backend == "vllm":
        from genlm.backend.llm import AsyncVirtualLM  # pragma: no cover

        model_cls = AsyncVirtualLM  # pragma: no cover
    elif backend == "hf":
        from genlm.backend.llm import AsyncTransformer

        model_cls = AsyncTransformer
    elif backend == "mlx":
        from genlm.backend.llm import AsyncMlxLM

        model_cls = AsyncMlxLM
    else:
        raise ValueError(
            f"Unknown backend: {backend}. Must be one of ['vllm', 'hf', 'mlx']"
        )  # pragma: no cover

    return model_cls.from_name(name, **kwargs)


class TokenMappings(NamedTuple):
    """
    Container for token mappings between bytes and tokens IDs in a language model.

    The `decode` and `encode` mappings are generally different from the `PromptedLLM` class (see notes on EOS token handling).
    """

    decode: list[bytes]  # token_id -> bytes
    encode: dict[bytes, int]  # bytes -> token_id
    eos_idxs: list[int]  # IDs of EOS tokens
    eos_tokens: list[bytes]  # EOS tokens
    potential_vocab: list[bytes]  # tokens in the potential's vocabulary

    @classmethod
    def create(cls, decode, eos_tokens):
        if len(set(eos_tokens)) != len(eos_tokens):
            raise ValueError("Duplicate eos tokens")
        encode = {x: i for i, x in enumerate(decode)}
        if not all(eos in encode for eos in eos_tokens):
            raise ValueError("EOS token not in language model vocabulary")
        eos_idxs = [encode[eos] for eos in eos_tokens]
        eos_tokens_set = set(eos_tokens)
        potential_vocab = [x for x in decode if x not in eos_tokens_set]
        return cls(
            decode=decode,
            encode=encode,
            eos_idxs=eos_idxs,
            eos_tokens=eos_tokens,
            potential_vocab=potential_vocab,
        )


class PromptedLLM(Potential):
    """A potential representing a language model conditioned on a fixed prompt prefix.

    `PromptedLLM`s operate on byte sequences.

    Notes on EOS Token Handling:\n
    - Tokens to treat as end-of-sequence tokens are specified via the `eos_tokens` argument.\n
    - These tokens are excluded from the potential's vocabulary and as such do not appear in the `vocab` attribute.\n
        This means they cannot appear in any input contexts to the potential nor in the output of `logw_next`. They can be used in the prompt however.\n
    - The log probability assigned to the `genlm.control`'s reserved `EOS` token is the sum of the log probabilities of all the specified EOS tokens.\n

    This class wraps an `AsyncLM` instance.
    """

    def __init__(
        self,
        llm,
        prompt_ids=None,
        eos_tokens=None,
        temperature=1.0,
        token_maps=None,
    ):
        """`
        Initializes the PromptedLLM potential.

        Args:
            llm (AsyncLM): The language model to use.
            prompt_ids (list[int], optional): Optional prompt to use as a prompt prefix for all input contexts.
                Must be a list of token IDs. Defaults to None. The prompt ids can be set post-init via `prompt` or `prompt_ids`.
            eos_tokens (list[bytes], optional): List of tokens to treat as end-of-sequence tokens.
                Defaults to the EOS token of the language model's tokenizer.
            temperature (float, optional): The temperature to apply to the language model's logits. Defaults to 1.
            token_maps (TokenMappings, optional): A precomputed mapping of tokens to token IDs with the potential's vocabulary.
                If provided, `eos_tokens` must not be provided. Defaults to None, which constructs a TokenMappings from the language model's byte vocabulary and the EOS tokens.
        """
        self.model = llm
        self.prompt_ids = prompt_ids or []
        self.temperature = temperature

        if token_maps is not None:
            if eos_tokens is not None:
                raise ValueError(
                    "eos_tokens must not be provided when token_maps is provided."
                )
            self.token_maps = token_maps
        else:
            self.token_maps = TokenMappings.create(
                decode=self.model.byte_vocab,
                eos_tokens=eos_tokens
                or [self.model.byte_vocab[self.model.tokenizer.eos_token_id]],
            )

        super().__init__(vocabulary=self.token_maps.potential_vocab)

    @classmethod
    def from_name(
        cls,
        name,
        backend=None,
        eos_tokens=None,
        prompt_ids=None,
        temperature=1.0,
        **kwargs,
    ):
        """Create a `PromptedLLM` from a Hugging Face model name.

        Args:
            name (str): Name of the model to load
            backend (str, optional): `AsyncLM` backend to use:\n
                * 'vllm' to instantiate an `AsyncVirtualLM`; ideal for GPU usage\n
                * 'hf' for an `AsyncTransformer`; ideal for CPU usage\n
                * 'mlx' for an `AsyncMlxLM`; ideal for Apple silicon usage\n
                * 'mock' for a `MockAsyncLM`; ideal for testing.\n
                Defaults to 'vllm' if CUDA is available, otherwise 'hf'.
            eos_tokens (list[bytes], optional): List of tokens to treat as end-of-sequence tokens.
                Defaults to the EOS token of the language model's tokenizer.
            prompt_ids (list[int], optional): Optional prompt to use as a prompt prefix for all input contexts.
                Must be a list of token IDs. Defaults to None. The prompt ids can be set post-init via `set_prompt_from_str` or `prompt_ids`.
            temperature (float, optional): The temperature to apply to the language model's logits. Defaults to 1.
            **kwargs (dict): Additional arguments passed to AsyncLM constructor

        Returns:
            (PromptedLLM): An instance of PromptedLLM
        """
        backend = backend or ("vllm" if torch.cuda.is_available() else "hf")
        model = load_model_by_name(name, backend=backend, **kwargs)
        return cls(
            model, prompt_ids=prompt_ids, eos_tokens=eos_tokens, temperature=temperature
        )

    @property
    def eos_tokens(self):
        return self.token_maps.eos_tokens

    @eos_tokens.setter
    def eos_tokens(self, value):
        raise ValueError(
            "Cannot reset eos_tokens after initialization. "
            "Use spawn_new_eos(new_eos_tokens) instead."
        )

    @property
    def prompt(self):
        """
        Get the current prompt as a list of byte sequences corresponding to the prompt token IDs.

        Returns:
            (list[bytes]|None): The current prompt as a list of bytes sequences or None if no prompt_ids are set.
        """
        if not self.prompt_ids:
            return  # pragma: no cover
        return [self.token_maps.decode[x] for x in self.prompt_ids]

    def set_prompt_from_str(self, prompt_str):
        """Set the fixed prompt from a string.

        Modifies `prompt_ids` to be the token IDs of the input prompt according to the language model's tokenizer.

        Args:
            prompt_str (str): The prompt to set.
        """
        # TODO: Handle race condition where prompt_ids reset concurrently.
        if not isinstance(prompt_str, str):
            raise ValueError(
                f"Prompt must a string got {type(prompt_str)}. "
                f"To set the prompt from a list of token IDs, use prompt_ids."
            )

        if prompt_str.endswith(" "):
            warnings.warn(
                "Prompt ends with whitespace, which may affect tokenization. "
                "Consider removing trailing whitespace.",
                stacklevel=2,
            )

        self.prompt_ids = self.model.tokenizer.encode(prompt_str)

    def encode_tokens(self, tokens):
        """Encode a list of byte tokens to a list of token IDs in
        the underlying language model's vocabulary.

        Args:
            tokens (list[bytes]): List of byte tokens to encode

        Returns:
            (list[int]): A list of token IDs corresponding to the input tokens.

        Raises:
            ValueError: If any token is not in the vocabulary
        """
        try:
            return [self.token_maps.encode[x] for x in tokens]
        except KeyError as e:
            raise ValueError(f"Token {e.args[0]} not in vocabulary") from e

    def decode_tokens(self, ids):
        """
        Decode a list of token IDs in the language model's vocabulary to a list of byte tokens.

        Args:
            ids (list[int]): A list of token IDs in the language model's vocabulary.

        Returns:
            (list[bytes]): A list of byte tokens corresponding to the input token IDs.
        """
        return [self.token_maps.decode[x] for x in ids]

    def tokenize(self, context_str):
        """Tokenize a string to a list of `bytes` objects, each corresponding to a token in the vocabulary.

        Uses the language model's tokenizer to map `context_str` to a list of token IDs, and then decodes the token IDs to bytes.

        Args:
            context_str (str): A string to encode

        Returns:
            (List[bytes]): A list of byte tokens corresponding to the input string.
        """
        return self.decode_tokens(self.model.tokenizer.encode(context_str))

    async def log_probability(self, context):
        """
        Compute the log probability of `context` given the prompt.

        Args:
            context (list[bytes]): A sequence of bytes tokens.

        Returns:
            (float): The log probability of `context`.
        """
        if not context:
            return 0

        context_ids = self.encode_tokens(context)
        return await self._log_probability(context_ids)

    async def _log_probability(self, context_ids):
        prefixes = [self.prompt_ids + context_ids[:i] for i in range(len(context_ids))]
        log_ps = self._maybe_temper(
            await self.model.batch_next_token_logprobs(prefixes)
        )
        target_ids = torch.tensor(context_ids, device=log_ps.device)
        with torch.no_grad():
            token_logprobs = torch.gather(log_ps, 1, target_ids.unsqueeze(1))
            total_logprob = token_logprobs.sum().item()

        return total_logprob

    def _maybe_temper(self, logps):
        if self.temperature == 1:
            return logps
        return torch.log_softmax(logps / self.temperature, dim=-1)

    async def prefix(self, context):
        """
        Compute the log probability of `context` given the prompt.

        Args:
            context (list[bytes]): A sequence of bytes tokens.

        Returns:
            (float): The log probability of `context`.
        """
        return await self.log_probability(context)

    async def complete(self, context):
        """
        Compute the log probability of `context` and the eos tokens given the prompt.

        If the model has multiple eos tokens, their probabilities will be summed.

        Args:
            context (list[bytes]): A sequence of bytes tokens.

        Returns:
            (float): The log probability of the context.
        """
        context_ids = self.encode_tokens(context)
        logp_context = await self._log_probability(context_ids)
        logp_next = self._maybe_temper(
            await self.model.next_token_logprobs(self.prompt_ids + context_ids)
        )
        logp_eos = torch.logsumexp(logp_next[self.token_maps.eos_idxs], dim=0).item()
        return logp_context + logp_eos

    def _process_logw_next(self, logw_next):
        """Process the log probabilities for the next tokens.

        This function rearranges the log probabilities such that the end-of-sequence (EOS) token's log probability
        is the sum of the log probabilities of `self.eos_tokens`.

        Args:
            logw_next (torch.tensor): The log probabilities for the next tokens.

        Returns:
            (LazyWeights): Processed log probabilities for the next tokens.
        """
        # This is ugly, but it's useful for all potentials to adhere to the convention
        # of keeping the EOS token at the end of the weights array.

        # Cache eos_idxs_tensor and non_eos_indices on first use
        if (
            not hasattr(self, "_eos_idxs_tensor")
            or not hasattr(self, "_non_eos_indices")
            or self._eos_idxs_tensor.device != logw_next.device
        ):
            self._eos_idxs_tensor = torch.tensor(
                self.token_maps.eos_idxs, device=logw_next.device
            )
            all_indices = torch.arange(
                len(self.token_maps.decode), device=logw_next.device
            )
            self._non_eos_indices = all_indices[
                ~torch.isin(all_indices, self._eos_idxs_tensor)
            ]

        logw_next = logw_next[: len(self.token_maps.decode)]
        logw_next = logw_next.log_softmax(dim=0)
        _logw_next = torch.full(
            (len(self.vocab) + 1,),
            float("-inf"),
            dtype=logw_next.dtype,
            device=logw_next.device,
        )
        _logw_next[: len(self.vocab)] = logw_next[self._non_eos_indices]

        # Special case: if only one EOS idx, just assign directly (avoids cost of logsumexp)
        if self._eos_idxs_tensor.numel() == 1:
            _logw_next[-1] = logw_next[self._eos_idxs_tensor]
        else:
            _logw_next[-1] = torch.logsumexp(logw_next[self._eos_idxs_tensor], dim=0)

        return self.make_lazy_weights(_logw_next.float().cpu().numpy())

    async def logw_next(self, context):
        """Get log probabilities for next tokens given the prompt and `context`.

        Args:
            context (List[bytes]): A sequence of bytes tokens.

        Returns:
            (LazyWeights): Log probabilities for next tokens and EOS.
        """
        logw_next = self._maybe_temper(
            await self.model.next_token_logprobs(
                self.prompt_ids + self.encode_tokens(context)
            )
        )
        return self._process_logw_next(logw_next)

    async def batch_logw_next(self, contexts):
        """Get log probabilities for next tokens given the prompt and `context`, for a batch of contexts.

        Args:
            contexts (list[list[bytes]]): A list of sequences of bytes tokens.

        Returns:
            (List[LazyWeights]): Log probabilities for next tokens and EOS for each context.
        """
        logw_nexts = self._maybe_temper(
            await self.model.batch_next_token_logprobs(
                [self.prompt_ids + self.encode_tokens(context) for context in contexts]
            )
        )
        return [self._process_logw_next(logw_next) for logw_next in logw_nexts]

    def __repr__(self):
        return f"PromptedLLM(prompt={self.prompt!r})"

    def spawn(self, prompt_ids=None, eos_tokens=None, temperature=None):
        """
        Spawn a new PromptedLLM.

        Args:
            prompt_ids (optional, list[int]): The prompt to use as a prompt prefix for all input contexts.
                Defaults to the same prompt_ids as `self`.
            eos_tokens (optional, list[bytes]): A list of tokens to treat as end-of-sequence tokens.
                Defaults to the same eos_tokens as `self`.
            temperature (optional, float): The temperature with which to rescale logprobs.
                Defaults to the same temperature as `self`.

        Returns:
            (PromptedLLM): A new PromptedLLM with the same prompt and eos tokens.

        Note:
            This is a shallow copy. The new PromptedLLM will share the underlying AsyncLM instance.
        """
        prompt_ids = prompt_ids if prompt_ids is not None else self.prompt_ids.copy()
        temperature = temperature if temperature is not None else self.temperature

        if (eos_tokens is None) or (eos_tokens == self.token_maps.eos_tokens):
            # If the eos tokens don't change, we don't need to recompute the token maps or vocabulary.
            return PromptedLLM(
                self.model,
                prompt_ids=prompt_ids,
                temperature=temperature,
                token_maps=self.token_maps,
            )

        return PromptedLLM(
            self.model,
            prompt_ids=prompt_ids,
            eos_tokens=eos_tokens,
            temperature=temperature,
        )

    def spawn_new_eos(self, eos_tokens):
        """
        Create a new PromptedLLM with a different set of end-of-sequence tokens.

        Args:
            eos_tokens (list[bytes]): A list of tokens to treat as end-of-sequence tokens.

        Returns:
            (PromptedLLM): A new PromptedLLM with the specified end-of-sequence tokens.
                The new model will have the same prompt_ids as `self`.
        """
        return self.spawn(eos_tokens=eos_tokens)

    def to_autobatched(self):
        raise ValueError("PromptedLLMs are autobatched by default.")



import pytest
from genlm.control.potential.built_in.json import (
    JsonSchema,
    TrivialSource,
    json_schema_parser,
    FixedSetParser,
    ARBITRARY_JSON,
    Incomplete,
    FLOAT_PARSER,
    chunk_to_complete_utf8,
    ParseError,
    FullValidatorJsonSchema,
    ParserPotential,
    StringSource,
    Input,
    FloatParser,
    WHITESPACE_PARSER,
    StringLiteralMatchingPatternParser,
    prune_to_validatable_prefix,
    PatriciaTrie,
)
from genlm.control.potential.streaming import AsyncSource
import json
from typing import Any
from dataclasses import dataclass
from hypothesis import given, strategies as st, assume, example, settings, reject
from hypothesis_jsonschema import from_schema
from jsonschema import SchemaError
import regex


@pytest.mark.asyncio
async def test_validates_a_list_of_integers():
    potential = JsonSchema({"type": "array", "items": {"type": "integer"}})

    assert await potential.prefix(b"[1,2,3") == 0.0
    assert await potential.prefix(b'["hello world"') == -float("inf")
    assert await potential.prefix(b"{") == -float("inf")


@pytest.mark.asyncio
async def test_rejects_as_prefix_when_no_valid_continuation():
    potential = JsonSchema({"type": "object"})

    assert await potential.prefix(b"}") == -float("inf")


@pytest.mark.asyncio
@pytest.mark.parametrize("schema", [{"type": "array", "items": {"type": "integer"}}])
@pytest.mark.parametrize(
    "context",
    [
        b"[1,2,3",
        b"[0]",
    ],
)
async def test_consistency_properties(schema, context):
    potential = JsonSchema(schema)
    await potential.assert_autoreg_fact(context)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "potential",
    [
        FullValidatorJsonSchema({"type": "array", "items": {"type": "integer"}}),
        ParserPotential(
            json_schema_parser({"type": "array", "items": {"type": "integer"}})
        ),
    ],
)
async def test_logw_next_has_results(potential):
    logs = await potential.logw_next(b"")
    assert logs[b"["[0]] == 0.0


@pytest.mark.asyncio
async def test_will_error_on_impossible_unicode_prefixes():
    potential = JsonSchema({"type": "object"})
    assert await potential.prefix([190] * 5) == -float("inf")


@st.composite
def basic_schema(draw):
    dtype = draw(
        st.sampled_from(
            [
                "null",
                "boolean",
                "integer",
                "number",
                "string",
                "enum",
            ]
        )
    )

    # Note: We do not currently implement patterns here, because it's too hard to do this
    # without hitting https://github.com/mrabarnett/mrab-regex/issues/571

    if dtype == "enum":
        return {
            "enum": draw(
                st.lists(
                    st.none()
                    | st.booleans()
                    | st.integers()
                    | st.floats(allow_nan=False, allow_infinity=False)
                    | st.text(),
                    min_size=1,
                    unique_by=lambda x: (type(x), x),
                )
            )
        }

    return {"type": dtype}


@st.composite
def composite_schema(draw, sub_schema):
    match draw(
        st.sampled_from(
            [
                "array",
                "anyOf",
                "object",
            ]
        )
    ):
        case "array":
            result = {"type": "array", "items": draw(sub_schema)}
            min_contains = draw(st.integers(0, 10))
            if min_contains > 0:
                result["minContains"] = min_contains
            if draw(st.booleans()):
                max_contains = draw(st.integers(min_contains, 20))
                result["maxContains"] = max_contains
            return result
        case "anyOf":
            return {"anyOf": draw(st.lists(sub_schema, min_size=2))}
        case "object":
            result = {"type": "object"}
            result["properties"] = draw(
                st.dictionaries(
                    st.from_regex("[A-Za-z0-9_]+"),
                    sub_schema,
                )
            )
            if result["properties"]:
                result["required"] = draw(
                    st.lists(st.sampled_from(sorted(result["properties"])), unique=True)
                )
            result["additionalProperties"] = draw(st.booleans())
            return result


json_schema = st.recursive(
    base=basic_schema(),
    extend=composite_schema,
)


@dataclass(frozen=True)
class JSONSchemaPotentialProblem:
    schema: Any
    document: bytes
    prefix: bytes

    @property
    def value(self):
        return json.loads(self.document)


@st.composite
def json_schema_potential_problem(draw):
    schema = draw(json_schema)
    value = draw(from_schema(schema))
    text = json.dumps(
        value,
        # Inverted so that this shrinks to True, as ascii-only
        # JSON is simpler.
        ensure_ascii=not draw(st.booleans()),
        # Similarly inverted so as to shrink to True, on the
        # theory that this means that if keys are out of
        # order in a shrunk example then it really matters.
        sort_keys=not draw(st.booleans()),
        indent=draw(st.one_of(st.none(), st.integers(0, 4))),
    )

    document = text.encode("utf-8")
    assert document
    assume(len(document) > 1)
    i = draw(st.integers(1, len(document) - 1))
    prefix = document[:i]
    assume(prefix.strip())

    return JSONSchemaPotentialProblem(schema=schema, document=document, prefix=prefix)


@pytest.mark.asyncio
@example(
    JSONSchemaPotentialProblem(
        schema={"type": "string"},
        document=b'"0\xc2\x80\xc2\x80"',
        prefix=b'"0\xc2\x80\xc2',
    )
)
@example(
    JSONSchemaPotentialProblem(
        schema={
            "type": "string",
        },
        document=b'"000000000\\u001f\xc2\x80\xc2\x80"',
        prefix=b'"000000000\\u001f\xc2\x80\xc2\x80',
    ),
)
@example(
    JSONSchemaPotentialProblem(
        schema={
            "type": "string",
        },
        document=b'"000\\u001f\xc2\x80\xc2\x80\xc2\x80"',
        prefix=b'"000\\u001f\xc2\x80\xc2\x80\xc2',
    ),
)
@example(
    JSONSchemaPotentialProblem(
        schema={"type": "array", "items": {"type": "integer"}},
        document=b"[0]",
        prefix=b"[",
    ),
)
@example(
    JSONSchemaPotentialProblem(
        schema={
            "anyOf": [
                {"type": "object", "properties": {}, "additionalProperties": False},
                {
                    "type": "object",
                    "properties": {"0": {"type": "null"}},
                    "required": ["0"],
                    "additionalProperties": False,
                },
            ]
        },
        document=b'{"0": null}',
        prefix=b"{",
    ),
)
@example(
    JSONSchemaPotentialProblem(
        schema={"type": "array", "items": {"enum": [543064729, 5]}},
        document=b"[543064729, 5]",
        prefix=b"[",
    ),
)
@example(
    JSONSchemaPotentialProblem(
        schema={
            "anyOf": [
                {"type": "null"},
                {"type": "boolean"},
                {"enum": [-10]},
                {"type": "integer"},
            ]
        },
        document=b"-1",
        prefix=b"-",
    )
)
@example(
    JSONSchemaPotentialProblem(
        schema={"anyOf": [{"type": "null"}, {"type": "boolean"}]},
        document=b"null",
        prefix=b"n",
    )
)
@example(
    JSONSchemaPotentialProblem(
        schema={
            "type": "object",
            "properties": {"0\x7f": {"type": "integer"}},
            "required": [],
            "additionalProperties": False,
        },
        document=b'{"0\\u007f": 0}',
        prefix=b"{",
    ),
)
@example(
    JSONSchemaPotentialProblem(
        schema={
            "type": "object",
            "properties": {"0": {"type": "null"}},
            "required": [],
            "additionalProperties": True,
        },
        document=b'{"": []}',
        prefix=b"{",
    ),
)
@example(
    JSONSchemaPotentialProblem(schema={"enum": [None, 10]}, document=b"10", prefix=b"1")
)
@example(
    JSONSchemaPotentialProblem(
        schema={
            "type": "array",
            "items": {
                "anyOf": [
                    {"type": "boolean"},
                    {"enum": [True]},
                    {"anyOf": [{"type": "null"}, {"type": "number"}]},
                ]
            },
        },
        document=b"[1.0000000000000001e+133]",
        prefix=b"[",
    )
)
@example(
    JSONSchemaPotentialProblem(
        schema={
            "type": "array",
            "items": {
                "anyOf": [{"enum": [0]}, {"type": "integer"}, {"type": "number"}]
            },
        },
        document=b"[0.5]",
        prefix=b"[",
    ),
)
@given(json_schema_potential_problem())
@settings(max_examples=25, deadline=None, report_multiple_bugs=False)
async def test_always_returns_correctly_on_valid_documents(problem):
    parser = json_schema_parser(problem.schema)
    await parser.parse_string(problem.document.decode("utf-8"))

    try:
        await parser.parse_string(problem.prefix.decode("utf-8"))
    except (Incomplete, UnicodeDecodeError):
        pass

    for potential in [
        ParserPotential(parser),
        FullValidatorJsonSchema(problem.schema),
    ]:
        assert await potential.prefix(problem.prefix) == 0.0
        assert await potential.prefix(problem.document) == 0.0
        if await potential.complete(problem.prefix) > -float("inf"):
            # This can sometimes happen because e.g. numeric literals can have
            # a prefix that is also a valid JSON value. We check here that the
            # prefix is actually valid JSON and if so allow it.
            json.loads(problem.prefix)
        assert await potential.complete(problem.document) == 0.0


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "format",
    [
        "ipv4",
        "date-time",
        "date",
        "date-time",
        # duration not present in Draft 7 which we're currently using.
        # "duration",
        "email",
        "hostname",
        "idn-hostname",
        "ipv4",
        "ipv6",
        "json-pointer",
        "relative-json-pointer",
        "time",
        "uri",
        "uri-reference",
    ],
)
async def test_validates_formats(format):
    potential = JsonSchema({"format": format, "type": "string"})
    assert await potential.complete(b'"hello world"') == -float("inf")


@pytest.mark.asyncio
async def test_validates_regex_format():
    potential = JsonSchema({"format": "regex", "type": "string"})
    assert await potential.complete(b'"["') == -float("inf")


@pytest.mark.asyncio
async def test_will_not_allow_nonsense_after_json():
    potential = JsonSchema({"type": "object"})
    assert await potential.prefix(b"{} hello world") == -float("inf")
    assert await potential.complete(b"{} hello world") == -float("inf")


@pytest.mark.asyncio
async def test_valid_prefix_for_schema_eg1():
    potential = JsonSchema(
        {
            "$schema": "http://json-schema.org/draft-04/schema#",
            "type": "array",
            "items": {
                "$schema": "http://json-schema.org/draft-04/schema#",
                "type": "object",
                "properties": {
                    "time": {"type": "string", "format": "date-time"},
                    "relayId": {"type": "string"},
                    "data": {
                        "type": "object",
                        "patternProperties": {
                            "^[0-9a-zA-Z_-]{1,255}$": {
                                "type": ["number", "string", "boolean"]
                            }
                        },
                        "additionalProperties": False,
                    },
                },
                "required": ["data"],
                "additionalProperties": False,
            },
        }
    )

    assert await potential.prefix(b"[{") == 0.0


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "ws",
    [
        b"\n\n\n",
        b"\n    \n",
    ],
)
async def test_forbids_weird_whitespace(ws):
    potential = JsonSchema({})
    assert await potential.prefix(ws) == -float("inf")


@pytest.mark.asyncio
async def test_rejects_as_prefix_when_invalid_key_has_been_started():
    potential = JsonSchema(
        {
            "type": "object",
            "properties": {
                "data": {
                    "type": "string",
                }
            },
            "required": ["data"],
            "additionalProperties": False,
        }
    )

    assert await potential.prefix(b'{"fo') == -float("inf")


@pytest.mark.asyncio
async def test_rejects_when_value_is_invalid_before_object_is_complete():
    potential = JsonSchema(
        {
            "type": "object",
            "properties": {
                "stuff": {
                    "type": "string",
                },
                "data": {
                    "type": "string",
                },
            },
            "additionalProperties": False,
        }
    )

    assert await potential.prefix(b'{"data": 1.0, ') == -float("inf")


@pytest.mark.asyncio
async def test_rejects_duplicated_key():
    potential = JsonSchema(
        {
            "type": "object",
        }
    )

    assert await potential.prefix(b'{"data": 1.0, "data"') == -float("inf")


@pytest.mark.asyncio
async def test_rejects_string_as_invalid_integer_before_complete():
    potential = JsonSchema(
        {
            "type": "integer",
        }
    )

    assert await potential.prefix(b'"') == -float("inf")


@pytest.mark.asyncio
async def test_accepts_basic_integer_list():
    potential = JsonSchema({"type": "array", "items": {"type": "integer"}})

    assert await potential.prefix(b"[0]") == 0.0
    assert await potential.complete(b"[0]") == 0.0

    logs = dict(await potential.logw_next(b"[0]"))
    for k, v in logs.items():
        # Forbid all ascii characters other than newline and space.
        if isinstance(k, int) and k < 128 and k not in b" \n":
            assert v == -float("inf")
    assert logs[potential.eos] == 0.0


@pytest.mark.asyncio
async def test_rejects_string_as_invalid_integer_inside_list():
    potential = JsonSchema({"type": "array", "items": {"type": "integer"}})

    assert await potential.prefix(b'["') == -float("inf")


@pytest.mark.asyncio
async def test_can_extend_zero_to_integer_list():
    schema = {"type": "array", "items": {"type": "integer"}}
    potential = JsonSchema(schema)
    assert await potential.prefix(b"[0,") == 0


@dataclass(frozen=True)
class SchemaAndDocument:
    schema: Any
    document: Any


@st.composite
def json_schema_and_document(draw):
    schema = draw(json_schema)
    document = draw(from_schema(schema))
    return SchemaAndDocument(schema, document)


@pytest.mark.asyncio
@settings(report_multiple_bugs=False, deadline=None)
@given(json_schema_and_document())
async def test_parser_for_schema_always_returns_document(sad):
    parser = json_schema_parser(sad.schema)
    text = json.dumps(sad.document)
    result = await parser.parse_string(text)
    assert result == sad.document


@pytest.mark.asyncio
@example(
    JSONSchemaPotentialProblem(schema={"type": "integer"}, document=b"-1", prefix=b"-"),
)
@example(
    JSONSchemaPotentialProblem(
        schema={"type": "string"}, document=b'"\xc2\x80"', prefix=b'"'
    )
)
@example(
    JSONSchemaPotentialProblem(
        schema={
            "type": "object",
            "properties": {
                "0": {"type": "null"},
                "0\x7f": {"type": "null"},
                "1": {"type": "null"},
            },
            "required": ["0", "0\x7f", "1"],
            "additionalProperties": False,
        },
        document=b'{"0": null, "0\x7f": null, "1": null}',
        prefix=b"{",
    ),
)
@example(
    JSONSchemaPotentialProblem(
        schema={"type": "array", "items": {"type": "number"}},
        document=b"[\n1.3941332551795901e+28\n]",
        prefix=b"[\n1.3941332551795901e+",
    ),
)
@settings(report_multiple_bugs=False, deadline=None)
@given(json_schema_potential_problem())
async def test_parser_for_schema_prefix_can_only_raise_incomplete(problem):
    parser = json_schema_parser(problem.schema)

    # Just to get coverage on the repr methods.
    repr(parser)

    whole_text = problem.document.decode("utf-8")
    result = await parser.parse_string(whole_text)
    assert result == problem.value

    try:
        text = problem.prefix.decode("utf-8")
    except UnicodeDecodeError:
        reject()
    try:
        await parser.parse_string(text)
    except Incomplete:
        pass


@st.composite
def json_object(draw):
    return draw(
        st.one_of(
            st.none(),
            st.booleans(),
            st.floats(allow_nan=False, allow_infinity=False),
            st.text(),
            st.lists(json_object()),
            st.dictionaries(st.text(), json_object()),
        )
    )


@pytest.mark.asyncio
@example(False)
@example(0.0)
@settings(report_multiple_bugs=False, deadline=None)
@given(json_object())
async def test_parser_for_arbitrary_json_can_parse_arbitrary_json(obj):
    text = json.dumps(obj)
    input = Input(TrivialSource(text))
    result = await ARBITRARY_JSON.parse(input)
    assert result == obj
    assert input.index == len(text)


@pytest.mark.asyncio
@settings(report_multiple_bugs=False, deadline=None)
@given(st.sets(st.text()))
async def test_correctly_handles_fixed_object_keys(keys):
    parser = json_schema_parser(
        {
            "type": "object",
            "properties": {key: {"type": "null"} for key in keys},
            "additionalProperties": False,
        }
    )

    x = {key: None for key in keys}
    s = json.dumps(x)
    result = await parser.parse_string(s)
    assert result == x


@pytest.mark.asyncio
async def test_float_parser_incomplete_literal():
    with pytest.raises(Incomplete):
        await FLOAT_PARSER.parse_string("0.")


@st.composite
def chunked_utf8(draw, base=None):
    if base is None:
        base = draw(st.text(min_size=1)).encode("utf-8")
    assume(len(base) > 1)
    offsets = draw(st.sets(st.integers(1, len(base) - 1)))
    offsets.update((0, len(base)))
    offsets = sorted(offsets)
    chunks = [base[u:v] for u, v in zip(offsets, offsets[1:])]
    assert b"".join(chunks) == base
    return chunks


@given(chunked_utf8())
@settings(report_multiple_bugs=False, deadline=None)
def test_utf8_chunking_always_splits_utf8(chunks):
    rechunked = list(chunk_to_complete_utf8(chunks))
    assert b"".join(rechunked) == b"".join(chunks)
    for chunk in rechunked:
        assert chunk
        chunk.decode("utf-8")


class BasicSource(AsyncSource):
    def __init__(self, blocks):
        self.__blocks = iter(blocks)

    async def more(self):
        try:
            return next(self.__blocks)
        except StopIteration:
            raise StopAsyncIteration()


@pytest.mark.asyncio
@given(chunked_utf8())
@settings(report_multiple_bugs=False, deadline=None)
async def test_utf8_chunking_always_splits_utf8_async(chunks):
    source = BasicSource(chunks)
    string_source = StringSource(source)

    buffer = bytearray()

    while True:
        try:
            chunk = await string_source.more()
        except StopAsyncIteration:
            break
        buffer.extend(chunk.encode("utf-8"))

    assert bytes(buffer) == b"".join(chunks)


@pytest.mark.asyncio
async def test_parser_raises_incomplete_on_empty_string():
    with pytest.raises(Incomplete):
        await FLOAT_PARSER.parse_string("")


@pytest.mark.asyncio
async def test_validates_a_list_of_integers_parser_only():
    parser = json_schema_parser({"type": "array", "items": {"type": "integer"}})

    with pytest.raises(Incomplete):
        await parser.parse_string("[1,2,3")

    with pytest.raises(ParseError):
        assert await parser.parse_string('["hello world"')

    with pytest.raises(ParseError):
        await parser.parse_string("{")


@pytest.mark.asyncio
async def test_can_calculate_many_prefixes():
    potential = JsonSchema({"type": "object"})

    for i in range(100):
        prefix = b'{ "' + str(i).encode("utf-8")
        pot = await potential.prefix(prefix)
        assert pot == 0.0


@pytest.mark.asyncio
async def test_raises_value_error_for_logw_next_of_bad_prefix():
    potential = JsonSchema({"type": "object"})
    with pytest.raises(ValueError):
        await potential.logw_next(b"[")


@pytest.mark.asyncio
async def test_json_validator_rejects_silly_whitespace():
    potential = FullValidatorJsonSchema({"type": "object"})
    assert await potential.prefix(b"\n\n\n") == -float("inf")
    assert await potential.complete(b"\n\n\n") == -float("inf")


@pytest.mark.asyncio
async def test_float_parser_can_continue_parsing_across_boundaries():
    source = BasicSource(["2", ".", "0", "1"])

    input = Input(source)

    parser = FloatParser()

    f = await input.parse(parser)

    assert f == 2.01


@dataclass(frozen=True)
class JSONSchemaPotentialProblemMulti:
    schema: Any
    document: bytes
    values: list[bytes]

    @property
    def value(self):
        return json.loads(self.document)


@st.composite
def json_schema_potential_problem_multi(draw):
    schema = draw(json_schema)
    value = draw(from_schema(schema))
    text = json.dumps(
        value,
        # Inverted so that this shrinks to True, as ascii-only
        # JSON is simpler.
        ensure_ascii=not draw(st.booleans()),
        # Similarly inverted so as to shrink to True, on the
        # theory that this means that if keys are out of
        # order in a shrunk example then it really matters.
        sort_keys=not draw(st.booleans()),
        indent=draw(st.one_of(st.none(), st.integers(0, 4))),
    )

    document = text.encode("utf-8")
    assert document
    assume(len(document) > 1)

    values = []

    for _ in range(draw(st.integers(1, 10))):
        offsets = draw(st.sets(st.integers(1, len(document) - 1), min_size=1))
        offsets = sorted(offsets)
        prefixes = [document[:v] for v in offsets]
        values.extend(prefixes)

    values = draw(st.permutations(values))
    values = values[: draw(st.integers(1, len(values)))]

    return JSONSchemaPotentialProblemMulti(
        schema=schema, document=document, values=values
    )


@pytest.mark.asyncio
async def test_can_reject_wrong_type_inside_any_of():
    schema = {
        "anyOf": [
            {
                "anyOf": [
                    {
                        "type": "object",
                    },
                ]
            },
        ]
    }

    parser = json_schema_parser(schema)
    potential = ParserPotential(parser)

    assert await potential.prefix(b'"') == -float("inf")


@pytest.mark.asyncio
async def test_can_reject_early_in_any_of():
    schema = {
        "anyOf": [
            {
                "type": "object",
                "properties": {
                    "a": {"type": "string"},
                },
                "required": ["a"],
            },
            {
                "type": "object",
                "properties": {
                    "a": {"type": "number"},
                },
                "required": ["a"],
            },
        ]
    }

    parser = json_schema_parser(schema)
    potential = ParserPotential(parser)

    assert await potential.prefix(b'"') == -float("inf")
    assert await potential.prefix(b'{"a":') == 0
    assert await potential.prefix(b'{"a": "') == 0
    assert await potential.prefix(b'{"a": 1') == 0
    assert await potential.prefix(b'{"a": {') == -float("inf")


@pytest.mark.asyncio
async def test_will_reject_invalid_unicode_at_end():
    potential = FullValidatorJsonSchema({"type": "object"})
    assert await potential.prefix(b"{ }\n\n    \xe2\x9d\x8d\xb0") == -float("inf")


def test_chunk_to_complete_utf8_will_error_on_invalid_unicode():
    with pytest.raises(UnicodeDecodeError):
        list(chunk_to_complete_utf8([b"{ }\n\n    \xe2\x9d\x8d\xb0"]))


@pytest.mark.asyncio
async def test_rejects_using_unicode_whitespace():
    pot = JsonSchema({"type": "object"})
    assert await pot.prefix("{ \u3000".encode("utf-8")) == -float("inf")


def test_chunking_immediately_rejects_invalid_utf8_bytes():
    def bad_bytes():
        yield b"\xc0"
        assert False

    with pytest.raises(UnicodeDecodeError):
        list(chunk_to_complete_utf8(bad_bytes()))


def test_chunking_bails_early_on_invalid_start_bytes():
    def bad_bytes():
        yield b"\xe3\x86\x8c\x80"
        assert False

    with pytest.raises(UnicodeDecodeError):
        list(chunk_to_complete_utf8(bad_bytes()))


@pytest.mark.asyncio
async def test_long_whitespace_at_start_is_rejected():
    validator = FullValidatorJsonSchema({"type": "object"})
    assert await validator.prefix(b"  ") == 0
    assert await validator.prefix(b"\n\n") == 0
    assert await validator.prefix(b"    ") == -float("inf")
    assert await validator.prefix(b"\n\n  ") == -float("inf")


@pytest.mark.asyncio
async def test_no_double_newline_after_start():
    potential = JsonSchema({"type": "object"})
    assert await potential.prefix(b"{\n\n") == -float("inf")
    assert await potential.prefix(b"{\n  \n") == -float("inf")


def test_repr_of_filter():
    assert "filter" in repr(WHITESPACE_PARSER)


@pytest.mark.asyncio
async def test_const_fails_fast():
    potential = ParserPotential(json_schema_parser({"const": False}))
    assert await potential.prefix(b" ") == 0
    assert await potential.prefix(b" f") == 0
    assert await potential.prefix(b" false") == 0
    assert await potential.prefix(b" n") == -float("inf")


@pytest.mark.asyncio
async def test_const_fails_fast_in_string_literals():
    potential = ParserPotential(json_schema_parser({"const": "Hello world"}))
    assert await potential.prefix(b" ") == 0
    assert await potential.prefix(b'"Hello') == 0
    assert await potential.prefix(b'"Hi') == -float("inf")


@pytest.mark.asyncio
async def test_const_in_object():
    potential = ParserPotential(
        json_schema_parser({"type": "object", "properties": {"foo": {"const": None}}})
    )

    assert await potential.prefix(b'{"foo": nu') == 0
    assert await potential.complete(b'{"foo": null}') == 0
    assert await potential.prefix(b'{"foo": f') == -float("inf")


def test_errors_on_bad_types():
    with pytest.raises(SchemaError):
        JsonSchema({"type": "float"})


@pytest.mark.asyncio
async def test_can_validate_patterns_in_incomplete_strings():
    parser = json_schema_parser({"type": "string", "pattern": "^[0-9]*$"})
    potential = ParserPotential(parser)

    for prefix in [
        '"',
        '"01234',
        '"01234"',
    ]:
        try:
            await parser.parse_string(prefix)
        except Incomplete:
            pass

        assert await potential.prefix(prefix.encode("utf-8")) == 0.0

    for prefix in [
        '"A',
        '"0000A',
    ]:
        with pytest.raises(ParseError):
            await parser.parse_string(prefix)

        assert await potential.prefix(prefix.encode("utf-8")) == -float("inf")


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "prefix",
    [
        "{",
        '{"id1": "0',
        '{"id1": "0",',
        '{"id1": "0",',
        '{"id1": "0", "id2": "A',
    ],
)
async def test_can_validate_patterns_in_incomplete_strings_in_objects_good_prefix(
    prefix,
):
    parser = json_schema_parser(
        {
            "type": "object",
            "properties": {
                "id1": {"type": "string", "pattern": "[0-9]+"},
                "id2": {"type": "string", "pattern": "[A-Z]+"},
            },
        }
    )
    potential = ParserPotential(parser)

    with pytest.raises(Incomplete):
        await parser.parse_string(prefix)

    assert await potential.prefix(prefix.encode("utf-8")) == 0.0


@pytest.mark.asyncio
@pytest.mark.parametrize("prefix", ['{"id2": "0", "id1": "A'])
async def test_can_validate_patterns_in_incomplete_strings_in_objects_bad_prefix(
    prefix,
):
    parser = json_schema_parser(
        {
            "type": "object",
            "properties": {
                "id1": {"type": "string", "pattern": "[0-9]+"},
                "id2": {"type": "string", "pattern": "[A-Z]+"},
            },
        }
    )
    potential = ParserPotential(parser)

    with pytest.raises(ParseError):
        await parser.parse_string(prefix)

    assert await potential.prefix(prefix.encode("utf-8")) == -float("inf")


@pytest.mark.asyncio
async def test_can_handle_incomplete_escape_sequences():
    parser = json_schema_parser({"type": "string", "pattern": ".*"})
    snowman = "\u2603"

    json_snowman = json.dumps(snowman, ensure_ascii=True)

    for i in range(len(json_snowman)):
        with pytest.raises(Incomplete):
            await parser.parse_string(json_snowman[:i])

    assert await parser.parse_string(json_snowman) == snowman


@pytest.mark.asyncio
async def test_errors_if_string_only_matches_a_prefix():
    parser = json_schema_parser({"type": "string", "pattern": "cabbages"})

    cab = json.dumps("cab")

    with pytest.raises(ParseError):
        assert await parser.parse_string(cab)


@pytest.mark.asyncio
async def test_patterns_apply_if_matching_anywhere():
    schema = {"type": "string", "pattern": "\\.(mp4|avi|mov|wmv|flv)$"}
    parser = json_schema_parser(schema)
    await parser.parse_string('"0.mp4"')


@pytest.mark.asyncio
async def test_patterns_reject_non_strings():
    schema = {"type": "string", "pattern": "\\.(mp4|avi|mov|wmv|flv)$"}
    parser = json_schema_parser(schema)

    with pytest.raises(ParseError):
        await parser.parse_string("0")


@pytest.mark.asyncio
async def test_patterns_apply_if_match_before_end_of_string():
    schema = {"type": "string", "pattern": "<[^>]+>"}
    parser = json_schema_parser(schema)

    await parser.parse_string('"<0>0"')


# List of 100 regular expressions for common data validation scenarios.
# Generated by Claude rather than mined from real data..
COMMON_PATTERNS = [
    # Email validation
    r"^[a-zA-Z0-9._%+-]+@[a-zA-Z0-9.-]+\.[a-zA-Z]{2,}$",
    r"^[a-zA-Z0-9.!#$%&\'*+/=?^_`{|}~-]+@[a-zA-Z0-9](?:[a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?(?:\.[a-zA-Z0-9](?:[a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?)*$",
    # Phone numbers
    r"^\+?1?-?\.?\s?\(?[0-9]{3}\)?[\s.-]?[0-9]{3}[\s.-]?[0-9]{4}$",
    r"^(\+\d{1,3}[- ]?)?\d{10}$",
    r"^\(\d{3}\)\s?\d{3}-\d{4}$",
    r"^\d{3}-\d{3}-\d{4}$",
    r"^\+\d{1,3}\s\d{1,4}\s\d{1,4}\s\d{1,9}$",
    # URLs
    r"^https?://(?:[-\w.])+(?:\:[0-9]+)?(?:/(?:[\w/_.])*(?:\?(?:[\w&=%.]*))?(?:\#(?:[\w.]*))?)?$",
    r"^(https?|ftp)://[^\s/$.?#].[^\s]*$",
    r"^https?://[a-zA-Z0-9.-]+\.[a-zA-Z]{2,}(?:/.*)?$",
    # IP addresses
    r"^(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)$",
    r"^((25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)$",
    r"^([0-9a-fA-F]{1,4}:){7}[0-9a-fA-F]{1,4}$",  # IPv6
    # Credit card numbers
    r"^4[0-9]{12}(?:[0-9]{3})?$",  # Visa
    r"^5[1-5][0-9]{14}$",  # MasterCard
    r"^3[47][0-9]{13}$",  # American Express
    r"^3[0-9]{4,}$",  # Diners Club
    r"^6(?:011|5[0-9]{2})[0-9]{12}$",  # Discover
    # Dates
    r"^\d{4}-\d{2}-\d{2}$",  # YYYY-MM-DD
    r"^\d{2}/\d{2}/\d{4}$",  # MM/DD/YYYY
    r"^\d{2}-\d{2}-\d{4}$",  # MM-DD-YYYY
    r"^\d{1,2}/\d{1,2}/\d{4}$",  # M/D/YYYY
    r"^\d{4}/\d{2}/\d{2}$",  # YYYY/MM/DD
    # Time formats
    r"^([01]?[0-9]|2[0-3]):[0-5][0-9]$",  # HH:MM
    r"^([01]?[0-9]|2[0-3]):[0-5][0-9]:[0-5][0-9]$",  # HH:MM:SS
    r"^(1[0-2]|0?[1-9]):[0-5][0-9]\s?(AM|PM)$",  # 12-hour format
    # Passwords (various strength requirements)
    r"^(?=.*[a-z])(?=.*[A-Z])(?=.*\d)[a-zA-Z\d]{8,}$",  # Min 8 chars, uppercase, lowercase, digit
    r"^(?=.*[a-z])(?=.*[A-Z])(?=.*\d)(?=.*[@$!%*?&])[A-Za-z\d@$!%*?&]{8,}$",  # With special chars
    r"^.{8,}$",  # At least 8 characters
    r"^(?=.*[A-Za-z])(?=.*\d)[A-Za-z\d]{6,}$",  # At least 6 chars with letter and digit
    # Social Security Numbers
    r"^\d{3}-\d{2}-\d{4}$",
    r"^\d{9}$",
    # ZIP codes
    r"^\d{5}$",  # US ZIP
    r"^\d{5}-\d{4}$",  # US ZIP+4
    r"^[A-Za-z]\d[A-Za-z] \d[A-Za-z]\d$",  # Canadian postal code
    # UUIDs
    r"^[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12}$",
    r"^[0-9a-fA-F]{32}$",  # UUID without hyphens
    # Numbers and currencies
    r"^\d+$",  # Positive integers
    r"^-?\d+$",  # Integers
    r"^\d*\.?\d+$",  # Positive decimal numbers
    r"^-?\d*\.?\d+$",  # Decimal numbers
    r"^\$\d{1,3}(,\d{3})*(\.\d{2})?$",  # Currency format
    r"^\d{1,3}(,\d{3})*(\.\d{2})?$",  # Number with commas
    # Usernames
    r"^[a-zA-Z0-9_]{3,20}$",
    r"^[a-zA-Z][a-zA-Z0-9_.-]{2,19}$",
    r"^[a-zA-Z0-9_.-]{3,30}$",
    # Names
    r"^[a-zA-Z\s]{2,50}$",
    r"^[a-zA-Z\'\-\s]{1,50}$",
    r"^[A-Z][a-z]+(\s[A-Z][a-z]+)*$",  # Proper case names
    # Alphanumeric strings
    r"^[a-zA-Z0-9]+$",
    r"^[a-zA-Z0-9\s]+$",
    r"^[a-zA-Z0-9_-]+$",
    # File extensions
    r"\.(jpg|jpeg|png|gif|bmp)$",  # Images
    r"\.(pdf|doc|docx|txt|rtf)$",  # Documents
    r"\.(mp3|wav|flac|aac)$",  # Audio
    r"\.(mp4|avi|mov|wmv|flv)$",  # Video
    # HTML tags
    r"<[^>]+>",
    r"<\/?[a-zA-Z][a-zA-Z0-9]*[^<>]*>",
    # Hex colors
    r"^#([A-Fa-f0-9]{6}|[A-Fa-f0-9]{3})$",
    r"^#[0-9A-Fa-f]{6}$",
    # Base64
    r"^[A-Za-z0-9+/]*={0,2}$",
    # JSON strings (basic)
    r"^[\{\[].*[\}\]]$",
    # MAC addresses
    r"^([0-9A-Fa-f]{2}[:-]){5}([0-9A-Fa-f]{2})$",
    # Version numbers
    r"^\d+\.\d+\.\d+$",  # Semantic versioning
    r"^\d+\.\d+$",  # Major.minor
    # Stock symbols
    r"^[A-Z]{1,5}$",
    # License plates (US format examples)
    r"^[A-Z]{3}\d{4}$",
    r"^[A-Z]{2}\d{5}$",
    # Geographic coordinates
    r"^-?(([1-8]?\d(\.\d+)?|90(\.0+)?),-?((1[0-7]\d)|([1-9]?\d))(\.\d+)?|180(\.0+)?)$",  # Lat,Long
    # Binary strings
    r"^[01]+$",
    # Hashtags
    r"^#[a-zA-Z0-9_]+$",
    # Mention patterns
    r"^@[a-zA-Z0-9_]+$",
    # SQL injection patterns (for blacklisting)
    r".*(union|select|insert|delete|update|drop|create|alter|exec|execute).*",
    # XSS patterns (for blacklisting)
    r".*<script.*?>.*</script>.*",
    # Domain names
    r"^[a-zA-Z0-9]([a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?(\.[a-zA-Z0-9]([a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?)*$",
    # Port numbers
    r"^([1-9][0-9]{0,3}|[1-5][0-9]{4}|6[0-4][0-9]{3}|65[0-4][0-9]{2}|655[0-2][0-9]|6553[0-5])$",
    # ISBN (10 and 13 digit)
    r"^(?:ISBN(?:-1[03])?:? )?(?=[0-9X]{10}$|(?=(?:[0-9]+[- ]){3})[- 0-9X]{13}$|97[89][0-9]{10}$|(?=(?:[0-9]+[- ]){4})[- 0-9]{17}$)(?:97[89][- ]?)?[0-9]{1,5}[- ]?[0-9]+[- ]?[0-9]+[- ]?[0-9X]$",
    # US State abbreviations
    r"^(AL|AK|AZ|AR|CA|CO|CT|DE|FL|GA|HI|ID|IL|IN|IA|KS|KY|LA|ME|MD|MA|MI|MN|MS|MO|MT|NE|NV|NH|NJ|NM|NY|NC|ND|OH|OK|OR|PA|RI|SC|SD|TN|TX|UT|VT|VA|WA|WV|WI|WY)$",
    # Language codes (ISO 639-1)
    r"^[a-z]{2}$",
    # Country codes (ISO 3166-1 alpha-2)
    r"^[A-Z]{2}$",
    # MIME types
    r"^[a-zA-Z][a-zA-Z0-9][a-zA-Z0-9\!#\$&\-\^]*\/[a-zA-Z0-9][a-zA-Z0-9\!#\$&\-\^]*$",
    # JWT tokens (basic structure)
    r"^[A-Za-z0-9-_]+\.[A-Za-z0-9-_]+\.[A-Za-z0-9-_]*$",
    # MongoDB ObjectId
    r"^[0-9a-fA-F]{24}$",
    # Credit card expiry (MM/YY)
    r"^(0[1-9]|1[0-2])\/([0-9]{2})$",
    # CVV codes
    r"^[0-9]{3,4}$",
    # Routing numbers (US banks)
    r"^[0-9]{9}$",
    # IBAN (basic format)
    r"^[A-Z]{2}[0-9]{2}[A-Z0-9]{4}[0-9]{7}([A-Z0-9]?){0,16}$",
    # Twitter handles
    r"^@[A-Za-z0-9_]{1,15}$",
    # Slack channels
    r"^#[a-z0-9_-]+$",
    # Discord user tags
    r"^.{3,32}#[0-9]{4}$",
    # YouTube video IDs
    r"^[a-zA-Z0-9_-]{11}$",
    # Bitcoin addresses
    r"^[13][a-km-zA-HJ-NP-Z1-9]{25,34}$",
    # Ethereum addresses
    r"^0x[a-fA-F0-9]{40}$",
    # ISBN-10
    r"^(?:[0-9]{9}X|[0-9]{10})$",
    # ISBN-13
    r"^97[89][0-9]{10}$",
    # VAT numbers (EU format)
    r"^[A-Z]{2}[0-9A-Z]+$",
    # Strong password with all character types
    r'^(?=.*[a-z])(?=.*[A-Z])(?=.*\d)(?=.*[!@#$%^&*()_+\-=\[\]{};\':"\\|,.<>\/?])[A-Za-z\d!@#$%^&*()_+\-=\[\]{};\':"\\|,.<>\/?]{12,}$',
    # Weak passwords (for blacklisting)
    r"^(password|123456|qwerty|abc123|letmein|admin|welcome)$",
    # Only letters
    r"^[a-zA-Z]+$",
    # Only digits
    r"^[0-9]+$",
    # Whitespace validation
    r"^\S+$",  # No whitespace
    r"^[^\s]+(\s[^\s]+)*$",  # No leading/trailing/multiple whitespace
]

COMMON_PATTERNS.sort(key=lambda s: (len(s), s))


@pytest.mark.asyncio
@given(st.data())
async def test_can_always_validate_prefixes_of_pattern_matching_strings(data):
    pattern = data.draw(st.sampled_from(COMMON_PATTERNS))

    string = data.draw(st.from_regex(pattern))

    # To avoid hitting https://github.com/mrabarnett/mrab-regex/issues/571
    assume(regex.search(pattern, string) is not None)

    literal = json.dumps(string)

    parser = StringLiteralMatchingPatternParser(pattern)

    await parser.parse_string(literal)

    prefix = literal[: data.draw(st.integers(0, len(literal) - 1))]

    with pytest.raises(Incomplete):
        await parser.parse_string(prefix)


@pytest.mark.asyncio
@given(st.sampled_from(COMMON_PATTERNS), st.text(min_size=1))
async def test_will_reject_incomplete_bad_strings(pattern, text):
    assume(pattern[0] == "^")
    assume(not regex.match(pattern, text, partial=True))

    literal = json.dumps(text)

    parser = StringLiteralMatchingPatternParser(pattern)

    with pytest.raises(ParseError):
        await parser.parse_string(literal)

    with pytest.raises(ParseError):
        await parser.parse_string(literal[:-1])


@pytest.mark.asyncio
async def test_cutting_through_a_surrogate_pair():
    parser = StringLiteralMatchingPatternParser("^[A-Z]{2}\\d{5}$")
    string = json.dumps("AA𐒠0000")

    await parser.parse_string(string)

    for i in range(len(string)):
        with pytest.raises(Incomplete):
            await parser.parse_string(string[:i])


@pytest.mark.asyncio
async def test_reject_half_of_a_surrogate_pair():
    parser = StringLiteralMatchingPatternParser("^[A-Z]{2}\\d{5}$")
    string = json.dumps("AA\ud801")

    with pytest.raises(ParseError):
        await parser.parse_string(string)


@pytest.mark.asyncio
async def test_will_reject_invalid_json_string():
    parser = StringLiteralMatchingPatternParser(".*")

    with pytest.raises(ParseError):
        await parser.parse_string('"\\0"')


@pytest.mark.asyncio
async def test_will_not_reject_partial_string():
    parser = StringLiteralMatchingPatternParser(".*")

    with pytest.raises(Incomplete):
        await parser.parse_string('"\\')


@pytest.mark.asyncio
async def test_will_handle_string_split_midway_valid():
    parser = StringLiteralMatchingPatternParser('"*')

    assert await parser.parse(Input(BasicSource(['"\\', '""']))) == '"'


@pytest.mark.asyncio
async def test_will_reject_string_split_midway_invalid():
    parser = StringLiteralMatchingPatternParser('^"*$')

    with pytest.raises(ParseError):
        assert await parser.parse(Input(BasicSource(['"\\', '"A', '"'])))


@pytest.mark.asyncio
async def test_union_of_integer_and_number():
    parser = json_schema_parser(
        {
            "type": "array",
            "items": {"anyOf": [{"type": "integer"}, {"type": "number"}]},
        }
    )

    assert await parser.parse_string("[0]") == [0]
    assert await parser.parse_string("[0.0]") == [0.0]
    xs = await parser.parse_string("[0.0, 0, 0.0]")
    assert xs == [0.0, 0, 0.0]
    assert list(map(type, xs)) == [float, int, float]


@pytest.mark.asyncio
async def test_union_of_objects():
    schema = {
        "anyOf": [
            {"type": "object", "properties": {}, "additionalProperties": False},
            {
                "type": "object",
                "properties": {"0": {"type": "null"}},
                "required": ["0"],
                "additionalProperties": False,
            },
        ]
    }

    document = '{"0": null}'

    parser = json_schema_parser(schema)

    assert await parser.parse_string(document) == {"0": None}


@pytest.mark.asyncio
async def test_empty_object():
    schema = {"type": "object", "properties": {}, "additionalProperties": False}

    parser = json_schema_parser(schema)

    assert await parser.parse_string("{}") == dict()

    with pytest.raises(ParseError):
        await parser.parse_string('{"')


@pytest.mark.asyncio
async def test_empty_object_allows_keys():
    parser = json_schema_parser({"type": "object"})

    for i in range(100):
        with pytest.raises(Incomplete):
            await parser.parse_string('{ "' + str(i))


@pytest.mark.asyncio
async def test_enum_parsing():
    values = ["two", "twin"]
    parser = json_schema_parser({"enum": values})

    for v in values:
        s = json.dumps(v)
        await parser.parse_string(s)
        for i in range(len(s)):
            with pytest.raises(Incomplete):
                await parser.parse_string(s[:i])

    for s in [
        "2",
        '"q',
        "30",
        "twa",
    ]:
        with pytest.raises(ParseError):
            await parser.parse_string(s)


@pytest.mark.asyncio
async def test_single_value_enum():
    values = ["1"]
    parser = json_schema_parser({"enum": values})
    await parser.parse_string('"1"')

    with pytest.raises(ParseError):
        await parser.parse_string('"2')


@pytest.mark.asyncio
async def test_union_of_integer_and_number_with_e_notation():
    schema = {
        "type": "array",
        "items": {"anyOf": [{"type": "null"}, {"type": "integer"}, {"type": "number"}]},
    }

    parser = json_schema_parser(schema)

    await parser.parse_string("[1e-05]") == [1e-05]


@pytest.mark.asyncio
async def test_integer_accepts_only_positive_e_notation():
    schema = {
        "type": "array",
        "items": {"type": "integer"},
    }

    parser = json_schema_parser(schema)

    await parser.parse_string("[1e05]")

    with pytest.raises(ParseError):
        await parser.parse_string("[1e-05]")


@pytest.mark.asyncio
async def test_catches_error_before_close_of_string():
    schema = {
        "type": "object",
        "properties": {
            "location": {"type": "string", "pattern": r"^[A-Za-z\s\-\.]+,\s[A-Z]{2}$"}
        },
        "required": ["location"],
    }
    parser = json_schema_parser(schema)

    with pytest.raises(ParseError):
        await parser.parse_string('{"location" : "Miami, Flo')


@pytest.mark.asyncio
@example(
    contents=[""],
    test_strings=["0"],
)
@example(
    contents=["", "00"],
    test_strings=["0"],
)
@example(
    contents=["010"],
    test_strings=["00"],
)
@example(
    contents=["", "01"],
    test_strings=["00"],
)
@given(
    contents=st.lists(st.text(), unique=True, min_size=1),
    test_strings=st.lists(st.text(min_size=1), unique=True),
)
async def test_fixed_set_parser(contents, test_strings):
    parser = FixedSetParser(contents)

    for s in contents:
        input = Input(TrivialSource(s))
        result = await parser.parse(input)
        assert result == s
        assert input.index == len(s)

    for s in test_strings:
        try:
            t = await parser.parse_string(s)
        except ParseError:
            for c in contents:
                assert not s.startswith(c)
        except Incomplete:
            for c in contents:
                assert not s.startswith(c)
            assert any(c.startswith(s) for c in contents)
        else:
            assert t in contents
            assert s.startswith(t)
            for t2 in contents:
                if s.startswith(t2):
                    assert len(t2) <= len(t)


def test_fixed_set_parser_not_empty():
    with pytest.raises(ValueError):
        FixedSetParser(())


@pytest.mark.parametrize(
    "document,expected", [(b"1", b""), (b'["foo", b"ba', b'["foo",')]
)
def test_prune_to_validatable_prefix(document, expected):
    assert prune_to_validatable_prefix(document) == expected


@pytest.mark.asyncio
async def test_parser_with_empty_properties():
    parser = json_schema_parser({"type": "object", "properties": {}})

    await parser.parse_string('{"foo": 1}')


@pytest.mark.asyncio
async def test_whitespace_parser_rejects_unicode_whitespace():
    with pytest.raises(ParseError):
        await WHITESPACE_PARSER.parse_string("\u3000")


def test_trie_adding_prefix_of_existing():
    trie = PatriciaTrie(["foobar"])

    trie.add_string("foo")

    assert trie.root.prefix == "foo"
    assert trie.root.accepting

    assert trie.root.children["b"].prefix == "ar"
    assert trie.root.children["b"].accepting


@pytest.mark.asyncio
async def test_rejects_if_required_keys_are_missing():
    parser = json_schema_parser(
        {
            "type": "object",
            "properties": {"foo": {"type": "string"}, "bar": {"type": "number"}},
            "required": ["foo"],
        }
    )

    for incomplete in [
        "{}",
        '{"bar": 0.0}',
    ]:
        with pytest.raises(ParseError):
            await parser.parse_string(incomplete)

    assert await parser.parse_string('{"foo": "hello"}') == {"foo": "hello"}
